import logging
import queue
import threading
import time
from datetime import timedelta
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set

from balsam.schemas import JobState
from balsam.util import Process, SigHandler

from .util import Queue

if TYPE_CHECKING:
    from balsam._api.models import Job, Session  # noqa: F401
    from balsam.client import RESTClient

logger = logging.getLogger(__name__)


class SessionThread(threading.Thread):
    """
    Creates and maintains lease on a Session object by periodically pinging API in background thread
    """

    TICK_PERIOD = timedelta(minutes=1)

    def __init__(self, client: "RESTClient", site_id: int, scheduler_id: Optional[int] = None) -> None:
        super().__init__(daemon=True)
        assert self.daemon
        batch_job_id = None
        if scheduler_id is not None:
            try:
                batch_job = client.BatchJob.objects.get(
                    site_id=site_id,
                    scheduler_id=scheduler_id,
                )
                batch_job_id = batch_job.id
            except client.BatchJob.DoesNotExist:
                logger.warning(
                    f"Failed to lookup BatchJob with scheduler_id {scheduler_id}. "
                    "Reverting to a Session without a BatchJob. "
                    "The run will still work, but Jobs will not be associated with a BatchJob."
                )
        self.session = client.Session.objects.create(site_id=site_id, batch_job_id=batch_job_id)

    def run(self) -> None:
        while True:
            if self.session.id is not None:
                self.session.tick()
                logger.debug("Ticked session")
            time.sleep(self.TICK_PERIOD.total_seconds())


class FixedDepthJobSource(Process):
    """
    A background process maintains a queue of `prefetch_depth` jobs meeting
    the criteria below. Prefer this JobSource to hide API latency for
    high-throughput, when the number of Jobs to process far exceeds the
    number of available compute resources (e.g. 128 nodes handling 100k jobs).

    WARNING: When the number of Jobs becomes comparable to allocated
    resources, launchers using this JobSource may prefetch too much and
    prevent effective work-sharing (i.e. one launcher hogs all the jobs in
    its queue, leaving the other launchers empty-handed).
    """

    def __init__(
        self,
        client: "RESTClient",
        site_id: int,
        prefetch_depth: int,
        filter_tags: Optional[Dict[str, str]] = None,
        states: Set[str] = {"PREPROCESSED", "RESTART_READY"},
        serial_only: bool = False,
        max_wall_time_min: Optional[int] = None,
        max_nodes_per_job: Optional[int] = None,
        max_aggregate_nodes: Optional[float] = None,
        scheduler_id: Optional[int] = None,
        app_ids: Optional[Set[int]] = None,
    ) -> None:
        super().__init__()
        self.queue: "Queue[Job]" = Queue()
        self.prefetch_depth = prefetch_depth

        self.client = client
        self.site_id = site_id
        self.session_thread: Optional[SessionThread] = None
        self.session: Optional["Session"] = None
        self.filter_tags = {} if filter_tags is None else filter_tags
        self.app_ids = set() if app_ids is None else app_ids
        self.states = states
        self.serial_only = serial_only
        self.max_wall_time_min = max_wall_time_min
        self.max_nodes_per_job = max_nodes_per_job
        self.max_aggregate_nodes = max_aggregate_nodes
        self.scheduler_id = scheduler_id
        self.start_time = time.time()
        self.fetch_limit = 2048

    def get_jobs(self, max_num_jobs: int) -> List["Job"]:
        fetched = []
        for _ in range(max_num_jobs):
            try:
                job = self.queue.get_nowait()
                job.objects = self.client.Job.objects
                fetched.append(job)
            except queue.Empty:
                break
        return fetched

    def get(self, timeout: Optional[float] = None) -> "Job":
        job = self.queue.get(block=True, timeout=timeout)
        job.objects = self.client.Job.objects
        return job

    def _run(self) -> None:
        sig_handler = SigHandler()
        self.client.close_session()

        if self.session_thread is None:
            self.session_thread = SessionThread(
                client=self.client, site_id=self.site_id, scheduler_id=self.scheduler_id
            )
            self.session = self.session_thread.session
            self.session_thread.start()

        assert self.session is not None

        while not sig_handler.wait_until_exit(timeout=1):
            qsize = self.queue.qsize()
            fetch_count = max(0, self.prefetch_depth - qsize)
            fetch_count = min(fetch_count, self.fetch_limit)
            logger.debug(f"JobSource queue depth is currently {qsize}. Fetching {fetch_count} more")
            if fetch_count:
                params = self._get_acquire_parameters(fetch_count)
                jobs = self.session.acquire_jobs(**params)
                if jobs:
                    logger.debug(
                        f"Acquired from session {self.session.id} (batch_job_id {self.session.batch_job_id})"
                    )
                    logger.info(f"JobSource acquired {len(jobs)} jobs:")
                for job in jobs:
                    self.queue.put_nowait(job)
        logger.info("Signal: JobSource cancelling tick thread and deleting API Session")
        self.queue.cancel_join_thread()
        self.session.delete()
        logger.info("JobSource exit graceful")

    def _get_acquire_parameters(self, num_jobs: int) -> Dict[str, Any]:
        request_time: Optional[float]
        if self.max_wall_time_min:
            elapsed_min = (time.time() - self.start_time) / 60.0
            request_time = self.max_wall_time_min - elapsed_min
        else:
            request_time = None
        return dict(
            max_num_jobs=num_jobs,
            max_nodes_per_job=self.max_nodes_per_job,
            max_aggregate_nodes=self.max_aggregate_nodes,
            max_wall_time_min=request_time,
            serial_only=self.serial_only,
            filter_tags=self.filter_tags,
            states=self.states,
            app_ids=self.app_ids,
        )


class SynchronousJobSource(object):
    """
    In this JobSource, `get_jobs` invokes a blocking API call and introduces
    latency. However, it allows greater flexibility in launchers requesting
    *just enough* work for the available resources, and is therefore better
    suited for work-sharing between launchers when the number of tasks does
    not far exceed the available resources. (Example: if you want two 5-node
    launchers to effectively split 10 jobs that take 1 hour each, use this
    JobSource to ensure that no resources go unused!)
    """

    def __init__(
        self,
        client: "RESTClient",
        site_id: int,
        filter_tags: Optional[Dict[str, str]] = None,
        states: Set[JobState] = {JobState.preprocessed, JobState.restart_ready},
        serial_only: bool = False,
        max_wall_time_min: Optional[int] = None,
        scheduler_id: Optional[int] = None,
        app_ids: Optional[Set[int]] = None,
    ) -> None:
        self.client = client
        self.site_id = site_id
        self.filter_tags = {} if filter_tags is None else filter_tags
        self.app_ids = set() if app_ids is None else app_ids
        self.states = states
        self.serial_only = serial_only
        self.max_wall_time_min = max_wall_time_min
        self.start_time = time.time()

        self.scheduler_id = scheduler_id
        self.session_thread = SessionThread(client=self.client, site_id=self.site_id, scheduler_id=self.scheduler_id)
        self.session_thread.start()
        self.session = self.session_thread.session

    def start(self) -> None:
        pass

    def terminate(self) -> None:
        logger.info("Signal: JobSource cancelling tick thread and deleting API Session")
        self.session.delete()
        logger.info("JobSource exit graceful")

    def join(self) -> None:
        pass

    def get_jobs(
        self, max_num_jobs: int, max_nodes_per_job: Optional[int] = None, max_aggregate_nodes: Optional[float] = None
    ) -> List["Job"]:
        request_time: Optional[int]
        if self.max_wall_time_min:
            elapsed_min = (time.time() - self.start_time) / 60.0
            request_time = round(self.max_wall_time_min - elapsed_min)
        else:
            request_time = None
        jobs = self.session.acquire_jobs(
            max_num_jobs=max_num_jobs,
            max_nodes_per_job=max_nodes_per_job,
            max_aggregate_nodes=max_aggregate_nodes,
            max_wall_time_min=request_time,
            serial_only=self.serial_only,
            filter_tags=self.filter_tags,
            states=self.states,
            app_ids=self.app_ids,
        )
        if jobs:
            logger.debug(f"Acquired from session {self.session.id} (batch_job_id {self.session.batch_job_id})")
            for job in jobs:
                logger.debug(f"Acquired id {job.id}: batch_job_id {job.batch_job_id}")
        return jobs



import logging
import os
import queue
import sys
from contextlib import contextmanager
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Iterator, Optional, Union

from balsam._api.app import ApplicationDefinition, AppType
from balsam.schemas import DeserializeError, JobState, JobUpdate
from balsam.site import BulkStatusUpdater, FixedDepthJobSource
from balsam.util import Process, SigHandler

if TYPE_CHECKING:
    from balsam._api.models import Job
    from balsam.client import RESTClient

logger = logging.getLogger(__name__)
PathLike = Union[str, Path]


@contextmanager
def job_context(workdir: Path, stdout_filename: str) -> Iterator[None]:
    old_stdout = sys.stdout
    old_stderr = sys.stderr
    old_cwd = os.getcwd()
    try:
        os.chdir(workdir)
    except FileNotFoundError:
        workdir.mkdir(parents=True, exist_ok=True)
        os.chdir(workdir)

    try:
        with open(workdir.joinpath(stdout_filename), "a") as fp:
            sys.stdout = fp
            sys.stderr = fp
            yield
    finally:
        os.chdir(old_cwd)
        sys.stdout = old_stdout
        sys.stderr = old_stderr


def transition_state(app: ApplicationDefinition) -> None:
    assert app.job.state is not None
    transition_func = {
        JobState.staged_in: app.preprocess,
        JobState.run_done: app.postprocess,
        JobState.run_error: app.handle_error,
        JobState.run_timeout: app.handle_timeout,
    }[app.job.state]
    try:
        msg = f"Running {transition_func.__name__} for Job {app.job.id}"
        logger.debug(msg)
        sys.stdout.write(f"#BALSAM {msg}\n")
        transition_func()
    except Exception as exc:
        logger.exception(f"An exception occured in {transition_func}: marking Job {app.job.id} FAILED")
        app.job.state = JobState.failed
        app.job.state_data = {
            "message": f"An exception occured in {transition_func}",
            "exception": str(exc),
        }


def read_pyapp_result(app: ApplicationDefinition) -> None:
    if app._app_type != AppType.PY_FUNC:
        return
    if not Path("job.out").is_file():
        logger.warning(f"Missing job.out in {Path.cwd()}")
        return

    logger.debug(f"Scanning {app.job.workdir / 'job.out'} for python app result")

    return_line = None
    except_line = None
    ret_marker = "BALSAM-RETURN-VALUE"
    exc_marker = "BALSAM-EXCEPTION"
    with open("job.out") as fp:
        for line in fp:
            if ret_marker in line:
                return_line = line
                break
            elif exc_marker in line:
                except_line = line
                break

    if app.job._update_model is None:
        app.job._update_model = JobUpdate()

    if return_line:
        return_line = return_line[return_line.find(ret_marker) :]
        payload = return_line.split(None, 1)[1]
        app.job._update_model.serialized_return_value = payload
        app.job._update_model.serialized_exception = ""
        logger.debug(f"Set serialized_return_value for Job in {app.job.workdir}")
    elif except_line:
        except_line = except_line[except_line.find(exc_marker) :]
        payload = except_line.split(None, 1)[1]
        app.job._update_model.serialized_exception = payload
        app.job._update_model.serialized_return_value = ""
        logger.debug(f"Set serialized_exception for Job in {app.job.workdir}")


def run_lifecycle_hook(app: ApplicationDefinition) -> Dict[str, Any]:
    # Check job.state BEFORE running state transition:
    job = app.job
    if job.state in (JobState.run_done, JobState.run_error):
        read_pyapp_result(app)
    transition_state(app)

    update_data = job._update_model.dict(exclude_unset=True) if job._update_model else {}
    update_data["id"] = job.id
    update_data["state"] = job.state
    update_data["state_timestamp"] = datetime.utcnow()
    update_data["state_data"] = job.state_data
    return update_data


def handle_job(job: "Job", data_path: Path) -> Dict[str, Any]:
    try:
        app_cls = ApplicationDefinition.load_by_id(job.app_id)
    except DeserializeError as exc:
        logger.exception(f"Failed to load app {job.app_id} for Job(id={job.id}, workdir={job.workdir}).")
        job.state = JobState.failed
        return {
            "id": job.id,
            "state": JobState.failed,
            "state_timestamp": datetime.utcnow(),
            "state_data": {
                "message": f"An exception occured while loading the App {job.app_id}",
                "exception": str(exc),
            },
        }

    app = app_cls(job)
    workdir = job.resolve_workdir(data_path)
    with job_context(workdir, "balsam.log"):
        update_data = run_lifecycle_hook(app)
    return update_data


def run_worker(
    job_source: "FixedDepthJobSource",
    status_updater: "BulkStatusUpdater",
    data_path: PathLike,
) -> None:
    sig_handler = SigHandler()
    if ApplicationDefinition._client is not None:
        ApplicationDefinition._client.close_session()
    data_path = Path(data_path).resolve()

    while not sig_handler.is_set():
        try:
            job = job_source.get(timeout=1)
        except queue.Empty:
            continue
        else:
            update_data = handle_job(job, data_path)
            status_updater.put(**update_data)
            logger.debug(f"Job {job.id} advanced to {job.state}")

    logger.info("Signal: ProcessingWorker exit")


class ProcessingService(object):
    def __init__(
        self,
        client: "RESTClient",
        site_id: int,
        prefetch_depth: int,
        apps_path: Path,
        data_path: Path,
        filter_tags: Optional[Dict[str, str]] = None,
        num_workers: int = 5,
    ) -> None:
        self.site_id = site_id
        ApplicationDefinition._set_client(client)
        try:
            ApplicationDefinition.load_by_site(self.site_id)  # Warms the cache
        except DeserializeError as exc:
            logger.warning(
                f"At least one App registered at this Site failed to deserialize: {exc}. "
                "Jobs running with this App will FAIL.  Please fix the App classes and/or the "
                "Balsam Site environment, and double check that the Apps can load OK."
            )
        self.job_source = FixedDepthJobSource(
            client=client,
            site_id=site_id,
            prefetch_depth=prefetch_depth,
            filter_tags=filter_tags,
            states={"STAGED_IN", "RUN_DONE", "RUN_ERROR", "RUN_TIMEOUT"},
        )
        self.status_updater = BulkStatusUpdater(client)

        self.workers = [
            Process(
                target=run_worker,
                args=(
                    self.job_source,
                    self.status_updater,
                    data_path,
                ),
            )
            for _ in range(num_workers)
        ]
        self._started = False

    def start(self) -> None:
        if not self._started:
            self.status_updater.start()
            self.job_source.start()
            for worker in self.workers:
                worker.start()
            self._started = True

    def terminate(self) -> None:
        self.job_source.terminate()
        for worker in self.workers:
            worker.terminate()

    def join(self) -> None:
        for worker in self.workers:
            worker.join()
        # Wait til workers DONE before killing status_updater:
        self.status_updater.terminate()
        self.job_source.join()
        self.status_updater.join()



import time

from .app_run import SubprocessAppRun


class ThetaAprun(SubprocessAppRun):
    """
    https://www.alcf.anl.gov/support-center/theta/running-jobs-and-submission-scripts
    """

    def _pre_popen(self) -> None:
        time.sleep(0.01)

    def _build_cmdline(self) -> str:
        node_ids = [nid for nid in self._node_spec.node_ids]
        nid_str = ",".join(map(str, node_ids))
        cpu_affinity = self._launch_params.get("cpu_affinity", "none")
        if cpu_affinity not in ["none", "depth"]:
            cpu_affinity = "none"
        args = [
            "aprun",
            "-n",
            self.get_num_ranks(),
            "-N",
            self._ranks_per_node,
            "-L",
            nid_str,
            "-cc",
            cpu_affinity,
            "-d",
            self._threads_per_rank,
            "-j",
            self._threads_per_core,
            self._cmdline,
        ]
        return " ".join(str(arg) for arg in args)



from typing import Any, Dict, List

from fastapi import APIRouter, Depends, status
from sqlalchemy import orm

from balsam import schemas
from balsam.server import settings
from balsam.server.models import Job, crud, get_session
from balsam.server.pubsub import pubsub

from .filters import SessionQuery

router = APIRouter()
auth = settings.auth.get_auth_method()


@router.get("/", response_model=schemas.PaginatedSessionsOut)
def list(
    db: orm.Session = Depends(get_session),
    user: schemas.UserOut = Depends(auth),
    q: SessionQuery = Depends(SessionQuery),
) -> Dict[str, Any]:
    """List Job processing Sessions running under the user's Sites."""
    count, sessions = crud.sessions.fetch(db, owner=user, filterset=q)
    return {"count": count, "results": sessions}


@router.post("/", response_model=schemas.SessionOut, status_code=status.HTTP_201_CREATED)
def create(
    session: schemas.SessionCreate, db: orm.Session = Depends(get_session), user: schemas.UserOut = Depends(auth)
) -> schemas.SessionOut:
    """Create a new Session for acquiring Jobs to process."""
    created_session, expired_jobs, expiry_events = crud.sessions.create(db, owner=user, session=session)
    result = schemas.SessionOut.from_orm(created_session)

    result_jobs = [schemas.JobOut.from_orm(job) for job in expired_jobs]
    result_events = [schemas.LogEventOut.from_orm(e) for e in expiry_events]
    db.commit()

    pubsub.publish(user.id, "create", "session", result)
    pubsub.publish(user.id, "bulk-update", "job", result_jobs)
    pubsub.publish(user.id, "bulk-create", "event", result_events)
    return result


@router.post("/{session_id}", response_model=List[schemas.JobOut])
def acquire(
    session_id: int,
    spec: schemas.SessionAcquire,
    db: orm.Session = Depends(get_session),
    user: schemas.UserOut = Depends(auth),
) -> List[Job]:
    """Acquire Jobs using the given session_id."""
    acquired_jobs, expired_jobs, expiry_events = crud.sessions.acquire(
        db, owner=user, session_id=session_id, spec=spec
    )
    result_jobs = [schemas.JobOut.from_orm(job) for job in expired_jobs]
    result_events = [schemas.LogEventOut.from_orm(e) for e in expiry_events]
    db.commit()
    pubsub.publish(user.id, "bulk-update", "job", result_jobs)
    pubsub.publish(user.id, "bulk-create", "event", result_events)
    return acquired_jobs


@router.put("/{session_id}")
def tick(session_id: int, db: orm.Session = Depends(get_session), user: schemas.UserOut = Depends(auth)) -> None:
    """Send a heartbeat to extend the given Session by id."""
    ts, expired_jobs, events = crud.sessions.tick(db, owner=user, session_id=session_id)
    result = {"id": session_id, "heartbeat": ts}
    pubsub.publish(user.id, "update", "session", result)
    result_jobs = [schemas.JobOut.from_orm(job) for job in expired_jobs]
    result_events = [schemas.LogEventOut.from_orm(e) for e in events]
    db.commit()
    pubsub.publish(user.id, "bulk-update", "job", result_jobs)
    pubsub.publish(user.id, "bulk-create", "event", result_events)


@router.delete("/{session_id}", status_code=status.HTTP_204_NO_CONTENT)
def delete(session_id: int, db: orm.Session = Depends(get_session), user: schemas.UserOut = Depends(auth)) -> None:
    """Delete a Session, freeing associated Jobs."""
    crud.sessions.delete(db, owner=user, session_id=session_id)
    db.commit()
    pubsub.publish(user.id, "delete", "session", {"id": session_id})



from typing import Any, Dict

from fastapi import APIRouter, Depends
from sqlalchemy import orm

from balsam import schemas
from balsam.server import settings
from balsam.server.models import LogEvent, crud, get_session
from balsam.server.utils import Paginator

from .filters import EventLogQuery

router = APIRouter()
auth = settings.auth.get_auth_method()


@router.get("/", response_model=schemas.PaginatedLogEventOut)
def list(
    db: orm.Session = Depends(get_session),
    user: schemas.UserOut = Depends(auth),
    paginator: Paginator[LogEvent] = Depends(Paginator),
    q: EventLogQuery = Depends(EventLogQuery),
) -> Dict[str, Any]:
    """List events associated with the user's Jobs."""
    count, events = crud.events.fetch(db, owner=user, paginator=paginator, filterset=q)
    return {"count": count, "results": events}



import logging
import os
import sys
from types import ModuleType
from typing import Any, Optional, cast

import click

from balsam import __version__
from balsam.cmdline import app, job, login, scheduler, site

server: Optional[ModuleType]

logger = logging.getLogger("balsam.cmdline")
try:
    from balsam.cmdline import server
except ModuleNotFoundError as e:
    logger.debug(f"Balsam server not installed: {e}")
    server = None

# Monkey-patch make_default_short_help: cut off after first line
_old_shorthelp = click.utils.make_default_short_help


def _new_shorthelp(help: str, max_length: int = 45) -> str:
    help = help.lstrip().split("\n")[0]
    return str(_old_shorthelp(help, max_length))


click.utils.make_default_short_help = _new_shorthelp
click.core.make_default_short_help = _new_shorthelp  # type: ignore


@click.group()
@click.version_option(version=__version__)
def _main() -> None:
    """
    Balsam Command Line Interface.

    Each subcommand is recursively documented; use `--help` to get
    information for each balsam subcommand.
    """
    pass


def main() -> None:
    try:
        _main()
    except Exception as e:
        if os.environ.get("BALSAM_CLI_TRACEBACK"):
            raise
        click.echo(f"{str(e).strip()}")
        click.echo("    [Export BALSAM_CLI_TRACEBACK=1 to see a full stack trace]")
        sys.exit(1)


LOAD_COMMANDS = [
    login.login,
    login.register,
    site.site,
    app.app,
    job.job,
    scheduler.queue,
]
if server is not None:
    LOAD_COMMANDS.append(cast(Any, server).server)

for cmd in LOAD_COMMANDS:
    _main.add_command(cmd)

if __name__ == "__main__":
    main()



import logging
import logging.handlers
import sys
from pathlib import Path
from typing import Any, Optional, Union


def log_uncaught_exceptions(exctype: Any, value: Any, tb: Any) -> None:
    root_logger = logging.getLogger("balsam.server")
    root_logger.error(f"Uncaught Exception {exctype}: {value}", exc_info=(exctype, value, tb))


def setup_logging(log_dir: Optional[Path], log_level: Union[str, int]) -> None:
    logging.getLogger("balsam").handlers.clear()

    logger = logging.getLogger("balsam.server")
    logger.handlers.clear()
    format = "%(asctime)s.%(msecs)03d | %(process)d | %(levelname)s | %(name)s:%(lineno)s] %(message)s"
    datefmt = "%Y-%m-%d %H:%M:%S"
    formatter = logging.Formatter(format, datefmt=datefmt)

    handler = (
        logging.handlers.RotatingFileHandler(
            filename=log_dir / "server-balsam.log",
            maxBytes=int(32 * 1e6),
            backupCount=3,
        )
        if log_dir
        else logging.StreamHandler()
    )
    handler.setFormatter(formatter)
    logger.setLevel(log_level)
    logger.addHandler(handler)
    sys.excepthook = log_uncaught_exceptions



from logging import getLogger
from typing import TYPE_CHECKING, Any, Dict, List

from pydantic import BaseModel, validator

from balsam._api.models import Job

if TYPE_CHECKING:
    from balsam.platform.compute_node import ComputeNode

logger = getLogger(__name__)


class InsufficientResources(Exception):
    pass


class NodeSpec(BaseModel):
    node_ids: List[str]
    hostnames: List[str]
    cpu_ids: List[List[int]] = []
    gpu_ids: List[List[str]] = []

    @validator("hostnames")
    def hostnames_len(cls, v: List[str], values: Dict[str, Any]) -> List[str]:
        if len(values["node_ids"]) != len(v):
            raise ValueError("Must provide same number of node_ids as hostnames")
        return v

    @validator("cpu_ids", always=True)
    def cpu_ids_len(cls, v: List[List[int]], values: Dict[str, Any]) -> List[List[int]]:
        if not v:
            v = [[] for _ in range(len(values["node_ids"]))]
        elif len(values["node_ids"]) != len(v):
            raise ValueError("Must provide same number of cpu_id lists")
        return v

    @validator("gpu_ids", always=True)
    def gpu_ids_len(cls, v: List[List[str]], values: Dict[str, Any]) -> List[List[str]]:
        if not v:
            v = [[] for _ in range(len(values["node_ids"]))]
        elif len(values["node_ids"]) != len(v):
            raise ValueError("Must provide same number of gpu_id lists")
        return v


class NodeManager:
    def __init__(self, node_list: "List[ComputeNode]", allow_node_packing: bool = True) -> None:
        self.nodes = node_list
        self.job_node_map: Dict[int, List[int]] = {}
        self.allow_node_packing = allow_node_packing

    def _assign_single_node(self, job_id: int, num_cpus: int, num_gpus: int, node_occupancy: float) -> NodeSpec:
        if not self.allow_node_packing:
            node_occupancy = 1.0
        for node_idx, node in enumerate(self.nodes):
            if node.check_fit(num_cpus, num_gpus, node_occupancy):
                spec = node.assign(job_id, num_cpus, num_gpus, node_occupancy)
                self.job_node_map[job_id] = [node_idx]
                return NodeSpec(
                    node_ids=[node.node_id],
                    hostnames=[node.hostname],
                    cpu_ids=[spec["cpu_ids"]],
                    gpu_ids=[spec["gpu_ids"]],
                )
        raise InsufficientResources

    def _assign_multi_node(self, job_id: int, num_nodes: int) -> NodeSpec:
        assigned_nodes = []
        assigned_idxs = []
        for node_idx, node in enumerate(self.nodes):
            if node.check_fit(num_cpus=0, num_gpus=0, occupancy=1.0):
                assigned_nodes.append(node)
                assigned_idxs.append(node_idx)
            if len(assigned_nodes) == num_nodes:
                break
        else:
            raise InsufficientResources
        node_ids, hostnames = [], []
        for node in assigned_nodes:
            node.assign(job_id, num_cpus=0, num_gpus=0, occupancy=1.0)
            node_ids.append(node.node_id)
            hostnames.append(node.hostname)
        self.job_node_map[job_id] = assigned_idxs
        return NodeSpec(node_ids=node_ids, hostnames=hostnames)

    def count_empty_nodes(self) -> int:
        return len([node for node in self.nodes if node.occupancy == 0.0])

    def aggregate_free_nodes(self) -> float:
        return float(len(self.nodes)) - sum(n.occupancy for n in self.nodes)

    def assign(self, job: Job) -> NodeSpec:
        assert job.id is not None
        return self.assign_from_params(
            id=job.id,
            num_nodes=job.num_nodes,
            ranks_per_node=job.ranks_per_node,
            threads_per_rank=job.threads_per_rank,
            threads_per_core=job.threads_per_core,
            gpus_per_rank=job.gpus_per_rank,
            node_occupancy=1.0 / job.node_packing_count,
        )

    def assign_from_params(
        self,
        id: int,
        num_nodes: int,
        ranks_per_node: int,
        threads_per_rank: int,
        threads_per_core: int,
        gpus_per_rank: float,
        node_occupancy: float,
        **kwargs: Any,
    ) -> NodeSpec:
        if num_nodes > 1:
            return self._assign_multi_node(id, num_nodes)
        num_cpus = max(1, int(ranks_per_node * threads_per_rank // threads_per_core))
        num_gpus = int(ranks_per_node * gpus_per_rank)
        return self._assign_single_node(id, num_cpus, num_gpus, node_occupancy)

    def free(self, job_id: int) -> None:
        node_idxs = self.job_node_map.pop(job_id)
        for idx in node_idxs:
            node = self.nodes[idx]
            node.free(job_id)



import logging
from collections import defaultdict
from itertools import islice
from math import ceil
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Tuple, cast
from urllib.parse import urlparse

from balsam.platform.transfer import TaskInfo, TransferInterface, TransferRetryableError, TransferSubmitError
from balsam.schemas import TransferItemState

from .service_base import BalsamService

if TYPE_CHECKING:
    from balsam._api.models import TransferItem, TransferItemQuery
    from balsam.client import RESTClient

logger = logging.getLogger(__name__)


def resolve_stage_in_paths(item: "TransferItem", workdir: Path) -> Tuple[Path, Path, bool]:
    src = item.remote_path
    dest = workdir.joinpath(item.local_path)
    if dest == workdir and not item.recursive:
        dest = dest.joinpath(src.name)
    if dest.parent != workdir:
        dest.parent.mkdir(parents=True, exist_ok=True)
    return src, dest, item.recursive


def resolve_stage_out_paths(item: "TransferItem", workdir: Path) -> Tuple[Path, Path, bool]:
    src = workdir.joinpath(item.local_path)
    dest = item.remote_path
    return src, dest, item.recursive


class TransferService(BalsamService):
    def __init__(
        self,
        client: "RESTClient",
        site_id: int,
        data_path: Path,
        transfer_interfaces: Dict[str, TransferInterface],
        transfer_locations: Dict[str, str],
        max_concurrent_transfers: int,
        transfer_batch_size: int,
        num_items_query_limit: int,
        service_period: int,
    ) -> None:
        super().__init__(client=client, service_period=service_period)
        self.site_id = site_id
        self.data_path = data_path
        self.transfer_locations = transfer_locations
        self.max_concurrent_transfers = max_concurrent_transfers
        self.transfer_interfaces = transfer_interfaces
        self.transfer_batch_size = transfer_batch_size
        self.num_items_query_limit = num_items_query_limit

    @staticmethod
    def build_task_map(transfer_items: Iterable["TransferItem"]) -> "defaultdict[str, List[TransferItem]]":
        task_map = defaultdict(list)
        for item in transfer_items:
            task_map[item.task_id].append(item)
        return task_map

    def update_transfers(self, transfer_items: List["TransferItem"], task: TaskInfo) -> None:
        current_state = cast(TransferItemState, task.state)
        if all(item.state == current_state for item in transfer_items):
            logger.debug(
                f"No change in {len(transfer_items)} TransferItems: "
                f"already in state {task.state} for task {task.task_id}"
            )
            return
        for item in transfer_items:
            item.state = current_state
            item.transfer_info = {**item.transfer_info, **task.info}
        self.client.TransferItem.objects.bulk_update(transfer_items)
        logger.info(f"Updated {len(transfer_items)} TransferItems " f"with task {task.task_id} to state {task.state}")

    @staticmethod
    def batch_iter(li: List[Any], bsize: int) -> Iterable[List[Any]]:
        nbatches = ceil(len(li) / bsize)
        for i in range(nbatches):
            yield li[i * bsize : (i + 1) * bsize]

    def item_batch_iter(self, items: "TransferItemQuery") -> Iterable[List["TransferItem"]]:
        transfer_candidates = defaultdict(list)
        for item in items[: self.num_items_query_limit]:
            transfer_candidates[(item.direction, item.location_alias)].append(item)
        task_keys = sorted(
            transfer_candidates.keys(),
            key=lambda k: len(transfer_candidates[k]),
            reverse=True,
        )
        for key in task_keys:
            key_items = transfer_candidates[key]
            for batch in self.batch_iter(key_items, bsize=self.transfer_batch_size):
                yield batch

    @staticmethod
    def error_items(items: List["TransferItem"], msg: str) -> None:
        logger.warning(msg)
        for item in items:
            item.state = TransferItemState.error
            item.transfer_info = {**item.transfer_info, "error": msg}

    def get_transfer_loc(self, loc_alias: str) -> Tuple[TransferInterface, str]:
        """Lookup trusted remote datastore and transfer method"""
        transfer_location = self.transfer_locations.get(loc_alias)
        if transfer_location is None:
            raise ValueError(f"Invalid location alias: {loc_alias}")
        parsed = urlparse(transfer_location)
        protocol = parsed.scheme
        remote_loc = parsed.netloc
        if protocol not in self.transfer_interfaces:
            raise ValueError(f"Unsupported Transfer protocol: {protocol}")
        transfer_interface = self.transfer_interfaces[protocol]
        return transfer_interface, remote_loc

    def get_workdirs_by_id(self, batch: List["TransferItem"]) -> Dict[int, Path]:
        """Build map of job workdirs, ensuring existence"""
        job_ids = list(set(item.job_id for item in batch))
        jobs = {job.id: job for job in self.client.Job.objects.filter(id=job_ids)}
        if len(jobs) < len(job_ids):
            raise RuntimeError("Could not find all Jobs for TransferItems")
        workdirs = {}
        for job_id, job in jobs.items():
            workdir = Path(self.data_path).joinpath(job.workdir).resolve()
            workdir.mkdir(parents=True, exist_ok=True)
            assert job_id is not None
            workdirs[job_id] = workdir
        return workdirs

    def submit_task(self, batch: List["TransferItem"]) -> None:
        loc_alias = batch[0].location_alias
        direction = batch[0].direction

        if direction not in ["in", "out"]:
            raise ValueError("direction must be in or out")
        if not all(item.location_alias == loc_alias for item in batch):
            raise ValueError("All items in batch must have same location_alias")
        if not all(item.direction == direction for item in batch):
            raise ValueError("All items in batch must have same direction")

        try:
            transfer_interface, remote_loc = self.get_transfer_loc(loc_alias)
        except ValueError as exc:
            return self.error_items(batch, str(exc))

        workdirs = self.get_workdirs_by_id(batch)
        path_resolver = resolve_stage_in_paths if direction == "in" else resolve_stage_out_paths
        transfer_paths = [path_resolver(item, workdirs[item.job_id]) for item in batch]

        try:
            task_id = transfer_interface.submit_task(remote_loc, direction, transfer_paths)
        except TransferSubmitError as exc:
            return self.error_items(batch, f"TransferSubmitError: {exc}")
        except TransferRetryableError as exc:
            logger.error(f"Non-fatal submit error: {exc}")
            return
        for item in batch:
            item.task_id = str(task_id)
            item.state = TransferItemState.active

    def run_cycle(self) -> None:
        TransferItem = self.client.TransferItem
        pending_submit = TransferItem.objects.filter(site_id=self.site_id, state={TransferItemState.pending})
        in_flight = TransferItem.objects.filter(
            site_id=self.site_id, state={TransferItemState.active, TransferItemState.inactive}
        )

        task_map = self.build_task_map(in_flight)
        task_ids: List[str] = list(task_map.keys())
        num_active_tasks = len(task_map)

        try:
            tasks = TransferInterface.poll_tasks(task_ids)
        except TransferRetryableError as exc:
            logger.error(f"Non-fatal error in poll_tasks: {exc}")
            return

        for task in tasks:
            items = task_map[task.task_id]
            self.update_transfers(items, task)

        max_new_tasks = max(0, self.max_concurrent_transfers - num_active_tasks)
        submit_batches = islice(self.item_batch_iter(pending_submit), max_new_tasks)
        for batch in submit_batches:
            if batch:
                self.submit_task(batch)
                TransferItem.objects.bulk_update(batch)

    def cleanup(self) -> None:
        pass



import logging
import os
import subprocess
import tempfile
from collections import Counter, defaultdict
from datetime import datetime
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union

import click
import dateutil.parser

from .scheduler import (
    SchedulerBackfillWindow,
    SchedulerJobLog,
    SchedulerJobStatus,
    SchedulerSubmitError,
    SubprocessSchedulerInterface,
)

PathLike = Union[Path, str]

logger = logging.getLogger(__name__)


def parse_cobalt_time_minutes(t_str: str) -> int:
    try:
        H, M, S = map(int, t_str.split(":"))
    except ValueError:
        return 0
    else:
        return H * 60 + M + round(S / 60)


class CobaltScheduler(SubprocessSchedulerInterface):
    status_exe = "qstat"
    submit_exe = "qsub"
    delete_exe = "qdel"
    backfill_exe = "nodelist"

    # maps scheduler states to Balsam states
    _job_states = {
        "queued": "queued",
        "starting": "queued",
        "running": "running",
        "exiting": "running",
        "killing": "running",
        "user_hold": "queued",
        "dep_hold": "queued",
        "dep_fail": "submit_failed",
        "admin_hold": "queued",
        "finished": "finished",
        "failed": "submit_failed",
    }

    @staticmethod
    def _job_state_map(scheduler_state: str) -> str:
        return CobaltScheduler._job_states.get(scheduler_state, "unknown")

    # maps Balsam status fields to the scheduler fields
    # should be a comprehensive list of scheduler status fields
    _status_fields = {
        "scheduler_id": "JobID",
        "state": "State",
        "wall_time_min": "WallTime",
        "queue": "Queue",
        "num_nodes": "Nodes",
        "project": "Project",
        "time_remaining_min": "TimeRemaining",
        "queued_time_min": "QueuedTime",
    }

    # when reading these fields from the scheduler apply
    # these maps to the string extracted from the output
    @staticmethod
    def _status_field_map(balsam_field: str) -> Optional[Callable[[str], Any]]:
        status_field_map: Dict[str, Callable[[str], Any]] = {
            "scheduler_id": lambda id: int(id),
            "state": CobaltScheduler._job_state_map,
            "queue": lambda queue: str(queue),
            "num_nodes": lambda n: int(n),
            "time_remaining_min": parse_cobalt_time_minutes,
            "queued_time_min": parse_cobalt_time_minutes,
            "project": lambda project: str(project),
            "wall_time_min": parse_cobalt_time_minutes,
        }
        return status_field_map.get(balsam_field, None)

    # maps node list states to Balsam node states
    _node_states = {
        "busy": "busy",
        "idle": "idle",
        "cleanup-pending": "busy",
        "down": "busy",
        "allocated": "busy",
    }

    @staticmethod
    def _node_state_map(nodelist_state: str) -> str:
        try:
            return CobaltScheduler._node_states[nodelist_state]
        except KeyError:
            logger.warning("node state %s is not recognized", nodelist_state)
            return "unknown"

    # maps the Balsam status fields to the node list fields
    # should be a comprehensive list of node list fields
    _nodelist_fields = {
        "id": "Node_id",
        "name": "Name",
        "queues": "Queues",
        "state": "Status",
        "mem": "MCDRAM",
        "numa": "NUMA",
        "wall_time_min": "Backfill",
    }

    # when reading these fields from the scheduler apply
    # these maps to the string extracted from the output
    @staticmethod
    def _nodelist_field_map(balsam_field: str) -> Callable[[str], Any]:
        nodelist_field_map = {
            "id": lambda id: int(id),
            "state": CobaltScheduler._node_state_map,
            "queues": lambda x: x.split(":"),
            "wall_time_min": lambda x: parse_cobalt_time_minutes(x),
        }
        return nodelist_field_map.get(balsam_field, lambda x: x)

    @staticmethod
    def _get_envs() -> Dict[str, str]:
        env = {}
        fields = CobaltScheduler._status_fields.values()
        env["QSTAT_HEADER"] = ":".join(fields)
        return env

    @staticmethod
    def _render_submit_args(
        script_path: Union[Path, str], project: str, queue: str, num_nodes: int, wall_time_min: int, **kwargs: Any
    ) -> List[str]:
        args = [
            CobaltScheduler.submit_exe,
            "-O",
            Path(script_path).with_suffix("").name,
            "-A",
            project,
            "-q",
            queue,
            "-n",
            str(int(num_nodes)),
            "-t",
            str(int(wall_time_min)),
            str(script_path),
        ]
        return args

    @staticmethod
    def _render_status_args(project: Optional[str], user: Optional[str], queue: Optional[str]) -> List[str]:
        os.environ.update(CobaltScheduler._get_envs())
        args = [CobaltScheduler.status_exe]
        if user is not None:
            args += ["-u", user]
        if project is not None:
            args += ["-A", project]
        if queue is not None:
            args += ["-q", queue]
        return args

    @staticmethod
    def _render_delete_args(job_id: Union[int, str]) -> List[str]:
        return [CobaltScheduler.delete_exe, str(job_id)]

    @staticmethod
    def _render_backfill_args() -> List[str]:
        return [CobaltScheduler.backfill_exe]

    @staticmethod
    def _parse_submit_output(submit_output: str) -> int:
        try:
            return int(submit_output)
        except ValueError:
            for line in submit_output.split("\n"):
                try:
                    return int(line)
                except ValueError:
                    continue
            else:
                raise SchedulerSubmitError(f"Failed to parse Job ID from output:\n{submit_output}")

    @staticmethod
    def _parse_status_output(raw_output: str) -> Dict[int, SchedulerJobStatus]:
        # TODO: this can be much more efficient with a compiled regex findall()
        status_dict = {}
        job_lines = raw_output.split("\n")[2:]
        for line in job_lines:
            if not line.strip():
                continue
            try:
                job_stat = CobaltScheduler._parse_status_line(line)
            except (ValueError, TypeError) as exc:
                logger.debug(f"Cannot parse job status: {line}\n{exc}")
                continue
            else:
                status_dict[job_stat.scheduler_id] = job_stat
        return status_dict

    @staticmethod
    def _parse_status_line(line: str) -> SchedulerJobStatus:
        fields = line.split()
        actual = len(fields)
        expected = len(CobaltScheduler._status_fields)
        if actual != expected:
            raise ValueError(f"Line has {actual} columns: expected {expected}:\n{fields}")

        status = {}
        for name, value in zip(CobaltScheduler._status_fields, fields):
            func = CobaltScheduler._status_field_map(name)
            if callable(func):
                status[name] = func(value)
        return SchedulerJobStatus(**status)

    @staticmethod
    def _parse_backfill_output(stdout: str) -> Dict[str, List[SchedulerBackfillWindow]]:
        raw_lines = stdout.strip().split("\n")
        nodelist = []
        node_lines = raw_lines[2:]
        for line in node_lines:
            try:
                line_dict = CobaltScheduler._parse_nodelist_line(line)
            except (ValueError, TypeError):
                logger.warning(f"Cannot parse nodelist line: {line}")
            else:
                if line_dict["wall_time_min"] > 0 and line_dict["state"] == "idle":
                    nodelist.append(line_dict)

        windows = CobaltScheduler._nodelist_to_backfill(nodelist)
        return windows

    @staticmethod
    def _parse_nodelist_line(line: str) -> Dict[str, Any]:
        fields = line.split()
        actual = len(fields)
        expected = len(CobaltScheduler._nodelist_fields)

        if actual != expected:
            raise ValueError(f"Line has {actual} columns: expected {expected}:\n{fields}")

        status = {}
        for name, value in zip(CobaltScheduler._nodelist_fields, fields):
            func = CobaltScheduler._nodelist_field_map(name)
            status[name] = func(value)
        return status

    @staticmethod
    def _nodelist_to_backfill(
        nodelist: List[Dict[str, Any]],
    ) -> Dict[str, List[SchedulerBackfillWindow]]:
        queue_bf_times = defaultdict(list)
        windows = defaultdict(list)

        for entry in nodelist:
            bf_time = entry["wall_time_min"]
            queues = entry["queues"]
            for queue in queues:
                queue_bf_times[queue].append(bf_time)

        for queue, bf_times in queue_bf_times.items():
            # Mapping {bf_time: num_nodes}
            bf_counter = Counter(bf_times)
            # {
            #    queue_name: [(bf_time1, num_nodes1), (bf_time2, num_nodes2), ...],
            # }
            # For each queue, sorted with longer times first
            queue_bf_times[queue] = sorted(bf_counter.items(), reverse=True)

        for queue, bf_list in queue_bf_times.items():
            running_total = 0
            for bf_time, num_nodes in bf_list:
                running_total += num_nodes
                windows[queue].append(SchedulerBackfillWindow(num_nodes=running_total, wall_time_min=bf_time))
        return windows

    @staticmethod
    def _parse_time(line: str) -> datetime:
        time_str = line[: line.find("(UTC)")]
        return dateutil.parser.parse(time_str)

    @staticmethod
    def _parse_logs(scheduler_id: int, job_script_path: Optional[PathLike]) -> SchedulerJobLog:
        if job_script_path is None:
            logger.warning("No job script path provided; cannot parse logs from scheduler_id alone")
            return SchedulerJobLog()
        logfile = Path(job_script_path).with_suffix(".cobaltlog")
        try:
            logger.info(f"Attempting to parse {logfile}")
            cobalt_log = logfile.read_text()
        except FileNotFoundError:
            logger.warning(f"Could not parse log: no file {logfile}")
            return SchedulerJobLog()

        lines = [line.strip() for line in cobalt_log.split("\n") if "(UTC)" in line]
        start_time = None
        for line in lines:
            if "COBALT_STARTTIME" in line:
                start_time = CobaltScheduler._parse_time(line)
                break
        if start_time:
            end_time = CobaltScheduler._parse_time(lines[-1])
            return SchedulerJobLog(start_time=start_time, end_time=end_time)
        logger.warning(f"Could not parse log: no line containing COBALT_STARTTIME in {logfile}")
        return SchedulerJobLog()

    @classmethod
    def discover_projects(cls) -> List[str]:
        """
        Get the user's allowed/preferred allocations
        """
        click.echo("Checking with Cobalt for your current allocations...")
        with tempfile.NamedTemporaryFile() as fp:
            os.chmod(fp.name, 0o777)
            proc = subprocess.run(
                f"qsub -t 35 -n 1 -A placeholder_null {fp.name}",
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                encoding="utf-8",
            )

        projects = []
        for line in proc.stdout.split("\n"):
            if "Projects available" in line:
                projects = line.split()[2:]
                break

        if not projects:
            projects = super().discover_projects()
        return projects



from passlib.context import CryptContext  # type: ignore

ctx = CryptContext(schemes=["bcrypt"], deprecated="auto")


def verify_password(plain: str, hashed: str) -> bool:
    return bool(ctx.verify(plain, hashed))


def get_hash(password: str) -> str:
    return str(ctx.hash(password))



import time

from .app_run import SubprocessAppRun


class SlurmRun(SubprocessAppRun):
    """
    https://slurm.schedmd.com/srun.html
    """

    def _build_cmdline(self) -> str:
        node_ids = [h for h in self._node_spec.hostnames]
        num_nodes = str(len(node_ids))
        args = [
            "srun",
            "-n",
            self.get_num_ranks(),
            "--ntasks-per-node",
            self._ranks_per_node,
            "--nodelist",
            ",".join(node_ids),
            "--nodes",
            num_nodes,
            "--cpus-per-task",
            self.get_cpus_per_rank(),
            self._cmdline,
        ]
        return " ".join(str(arg) for arg in args)

    def _pre_popen(self) -> None:
        time.sleep(0.01)



import json
import logging
import time
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, Tuple

import zmq

from balsam._api.app import ApplicationDefinition
from balsam.config import SiteConfig
from balsam.schemas import DeserializeError, JobState
from balsam.site import BulkStatusUpdater, FixedDepthJobSource
from balsam.site.launcher.util import countdown_timer_min
from balsam.util import SigHandler

if TYPE_CHECKING:
    from balsam._api.models import Job
    from balsam.platform.app_run import AppRun  # noqa: F401

logger = logging.getLogger(__name__)


class Master:
    def __init__(
        self,
        job_source: FixedDepthJobSource,
        status_updater: BulkStatusUpdater,
        wall_time_min: int,
        master_port: int,
        data_dir: Path,
        idle_ttl_sec: int,
        num_workers: int,
    ) -> None:
        self.job_source = job_source
        self.status_updater = status_updater
        self.data_dir = data_dir
        self.remaining_timer = countdown_timer_min(wall_time_min, delay_sec=0)
        self.idle_ttl_sec = idle_ttl_sec
        self.idle_time: Optional[float] = None
        self.active_ids: Set[int] = set()
        self.num_outstanding_jobs: int = 0
        self.occupancies: Dict[int, float] = {}
        self.num_workers = num_workers
        self.master_port = master_port

        next(self.remaining_timer)

        self.sig_handler = SigHandler()
        self.status_updater.start()
        self.job_source.start()
        logger.debug("Job source/status updater created")

    def job_to_dict(self, job: "Job") -> Dict[str, Any]:
        app_cls = ApplicationDefinition.load_by_id(job.app_id)
        app = app_cls(job)
        workdir = self.data_dir.joinpath(app.job.workdir).as_posix()

        preamble = app.shell_preamble()
        app_command = app.get_arg_str()
        environ_vars = app.get_environ_vars()
        occ = 1.0 / job.node_packing_count
        assert job.id is not None
        self.occupancies[job.id] = occ
        return dict(
            id=job.id,
            cwd=workdir,
            cmdline=app_command,
            preamble=preamble,
            node_occupancy=occ,
            envs=environ_vars,
            threads_per_rank=job.threads_per_rank,
            threads_per_core=job.threads_per_core,
            gpus_per_rank=job.gpus_per_rank,
        )

    def update_job_states(
        self, done_ids: List[int], error_logs: List[Tuple[int, int, str]], started_ids: List[int]
    ) -> None:
        now = datetime.utcnow()

        for id in done_ids:
            self.status_updater.put(id, JobState.run_done, state_timestamp=now)

        for (id, retcode, tail) in error_logs:
            self.status_updater.put(
                id, JobState.run_error, state_timestamp=now, state_data={"returncode": retcode, "error": tail}
            )

        for id in started_ids:
            self.status_updater.put(
                id, JobState.running, state_timestamp=now, state_data={"num_nodes": self.occupancies.pop(id, 1.0)}
            )

    def acquire_jobs(self, max_jobs: int) -> List[Dict[str, Any]]:
        next_jobs = self.job_source.get_jobs(max_jobs)
        new_job_specs = []
        for job in next_jobs:
            assert job.id is not None
            try:
                spec = self.job_to_dict(job)
            except DeserializeError as exc:
                logger.exception(f"Failed to deserialize for Job(id={job.id}, workdir={job.workdir}): {exc}")
                self.status_updater.put(
                    job.id,
                    state=JobState.failed,
                    state_timestamp=datetime.utcnow(),
                    state_data={
                        "message": "An exception occured while loading the app or parameters",
                        "exception": str(exc),
                    },
                )
            else:
                new_job_specs.append(spec)
        return new_job_specs

    def handle_request(self) -> None:
        msg = self.socket.recv_json()

        done_ids: List[int] = msg["done"]
        error_logs: List[Tuple[int, int, str]] = msg["error"]
        started_ids: List[int] = msg["started"]
        self.update_job_states(done_ids, error_logs, started_ids)

        finished_ids = set(done_ids) | set(log[0] for log in error_logs)
        self.active_ids |= set(started_ids)
        self.active_ids -= finished_ids
        self.num_outstanding_jobs -= len(finished_ids)

        src = msg["source"]
        max_jobs: int = msg["request_num_jobs"]
        logger.debug(f"Worker {src} requested {max_jobs} jobs")
        new_job_specs = self.acquire_jobs(max_jobs)

        self.socket.send_json({"new_jobs": new_job_specs})
        self.num_outstanding_jobs += len(new_job_specs)
        if new_job_specs:
            logger.debug(f"Sent {len(new_job_specs)} new jobs to {src}")

    def idle_check(self) -> None:
        if not self.status_updater.is_alive():
            logger.error("StatusUpdater is DOWN.  Aborting job!")
            self.sig_handler.set()

        if not self.num_outstanding_jobs:
            if self.idle_time is None:
                self.idle_time = time.time()
            if time.time() - self.idle_time > self.idle_ttl_sec:
                logger.info(f"Nothing to do for {self.idle_ttl_sec} seconds: quitting")
                self.sig_handler.set()
        else:
            self.idle_time = None

    def run(self) -> None:
        logger.debug("In master run")
        try:
            self.context = zmq.Context()  # type: ignore
            self.context.setsockopt(zmq.LINGER, 0)  # type: ignore
            self.socket = self.context.socket(zmq.REP)  # type: ignore
            self.socket.bind(f"tcp://*:{self.master_port}")
            logger.debug("Master ZMQ socket bound.")

            for remaining_minutes in self.remaining_timer:
                logger.debug(f"{remaining_minutes} minutes remaining")
                self.handle_request()
                self.idle_check()
                if self.sig_handler.is_set():
                    logger.info("Signal: master breaking main loop")
                    break
        except:  # noqa
            raise
        finally:
            self.shutdown()
            self.socket.setsockopt(zmq.LINGER, 0)
            self.socket.close(linger=0)
            self.context.term()  # type: ignore
            logger.info("shutdown done: ensemble master exit gracefully")

    def shutdown(self) -> None:
        now = datetime.utcnow()
        logger.info(f"Timing out {len(self.active_ids)} active runs")
        for id in self.active_ids:
            self.status_updater.put(
                id,
                JobState.run_timeout,
                state_timestamp=now,
            )

        logger.info("Terminating StatusUpdater...")
        # StatusUpdater needs to join first: stop the RUNNING updates
        self.status_updater.terminate()
        self.status_updater.join()
        logger.info("StatusUpdater has joined.")

        # We trigger JobSourceexit after StatusUpdater has joined to ensure
        # *all* Jobs get properly released
        logger.info("Terminating JobSource...")
        self.job_source.queue.cancel_join_thread()
        logger.debug("Called cancel join thread")
        self.job_source.terminate()
        logger.debug("Sent SIGTERM to jobsource")
        logger.debug("blocking on job_source.join")
        self.job_source.join()
        logger.info("JobSource has joined.")
        logger.info("Master sending exit message to all Workers")
        for _ in range(self.num_workers):
            self.socket.recv_json()
            self.socket.send_json({"exit": True})
        logger.info("All workers have received exit message. Quitting.")


def master_main(wall_time_min: int, master_port: int, log_filename: str, num_workers: int, filter_tags: str) -> None:
    site_config = SiteConfig()
    filter_tags_dict: Optional[Dict[str, str]] = json.loads(filter_tags)

    site_config.enable_logging("serial_mode", filename=log_filename + ".master")
    logger.debug("Launching master")

    ApplicationDefinition._set_client(site_config.client)
    try:
        ApplicationDefinition.load_by_site(site_config.site_id)  # Warms the cache
    except DeserializeError as exc:
        logger.warning(
            f"At least one App registered at this Site failed to deserialize: {exc}. "
            "Jobs running with this App will FAIL.  Please fix the App classes and/or the "
            "Balsam Site environment, and double check that the Apps can load OK."
        )

    launch_settings = site_config.settings.launcher
    node_cls = launch_settings.compute_node
    scheduler_id = node_cls.get_scheduler_id()
    job_source = FixedDepthJobSource(
        client=site_config.client,
        site_id=site_config.site_id,
        prefetch_depth=num_workers * launch_settings.serial_mode_prefetch_per_rank,
        filter_tags=filter_tags_dict,
        max_wall_time_min=wall_time_min,
        scheduler_id=scheduler_id,
        serial_only=True,
        max_nodes_per_job=1,
    )
    status_updater = BulkStatusUpdater(site_config.client)

    master = Master(
        job_source=job_source,
        status_updater=status_updater,
        wall_time_min=wall_time_min,
        master_port=int(master_port),
        data_dir=site_config.data_path,
        idle_ttl_sec=launch_settings.idle_ttl_sec,
        num_workers=num_workers,
    )
    master.run()



from typing import Any, Dict, List, Optional, Tuple, Union, cast

from sqlalchemy.orm import Query, Session

from balsam import schemas
from balsam.server import ValidationError, models
from balsam.server.routers.filters import BatchJobQuery
from balsam.server.utils import Paginator


def fetch(
    db: Session,
    owner: schemas.UserOut,
    paginator: Optional[Paginator[models.BatchJob]] = None,
    batch_job_id: Optional[int] = None,
    filterset: Optional[BatchJobQuery] = None,
) -> "Tuple[int, Union[List[models.BatchJob], Query[models.BatchJob]]]":
    qs: "Query[models.BatchJob]" = db.query(models.BatchJob).join(models.Site).filter(models.Site.owner_id == owner.id)  # type: ignore
    if filterset is not None:
        qs = filterset.apply_filters(qs)
    if batch_job_id is not None:
        qs = qs.filter(models.BatchJob.id == batch_job_id)
        return 1, [qs.one()]
    count = qs.group_by(models.BatchJob.id).count()
    assert paginator is not None
    batch_jobs = paginator.paginate(qs)
    return count, batch_jobs


def create(db: Session, owner: schemas.UserOut, batch_job: schemas.BatchJobCreate) -> models.BatchJob:
    site_id = (
        db.query(models.Site.id)
        .filter(models.Site.owner_id == owner.id, models.Site.id == batch_job.site_id)
        .scalar()  # type: ignore
    )
    if site_id is None:
        raise ValidationError(f"No site with ID {batch_job.site_id}")
    created_batch_job = models.BatchJob(state=schemas.BatchJobState.pending_submission, **batch_job.dict())
    db.add(created_batch_job)
    db.flush()
    return created_batch_job


def _update_fields(in_db: models.BatchJob, update_dict: Dict[str, Any]) -> None:
    state_update = update_dict.pop("state", in_db.state)
    for k, v in update_dict.items():
        setattr(in_db, k, v)
    if in_db.state == schemas.BatchJobState.pending_deletion:
        if state_update != schemas.BatchJobState.finished:
            return
    in_db.state = state_update


def update(
    db: Session, owner: schemas.UserOut, batch_job_id: int, batch_job: schemas.BatchJobUpdate
) -> models.BatchJob:
    in_db = cast(
        models.BatchJob,
        db.query(models.BatchJob)
        .join(models.Site)  # type: ignore
        .filter(models.Site.owner_id == owner.id)
        .filter(models.BatchJob.id == batch_job_id)
        .with_for_update(of=models.BatchJob, nowait=False, skip_locked=False)
        .one(),
    )
    _update_fields(in_db, batch_job.dict(exclude_unset=True))
    db.flush()
    return in_db


def bulk_update(
    db: Session, owner: schemas.UserOut, batch_jobs: List[schemas.BatchJobBulkUpdate]
) -> List[models.BatchJob]:
    update_map = {j.id: j.dict(exclude_unset=True) for j in batch_jobs}
    ids = set(update_map.keys())
    qs: "Query[models.BatchJob]" = (
        db.query(models.BatchJob)
        .join(models.Site)  # type: ignore
        .filter(models.Site.owner_id == owner.id)
        .filter(models.BatchJob.id.in_(ids))
        .order_by(models.BatchJob.id)
        .with_for_update(of=models.BatchJob)
    )
    db_jobs = {j.id: j for j in qs.all()}
    if len(db_jobs) < len(update_map):
        raise ValidationError("Could not find some Batch Job IDs")

    for id, update_dict in update_map.items():
        db_job = db_jobs[id]
        _update_fields(db_job, update_dict)

    db.flush()
    return list(db_jobs.values())


def delete(db: Session, owner: schemas.UserOut, batch_job_id: int) -> None:
    qs = (
        db.query(models.BatchJob)
        .join(models.Site)  # type: ignore
        .filter(models.Site.owner_id == owner.id)
        .filter(models.BatchJob.id == batch_job_id)
    )
    bjob = qs.one()
    db.query(models.BatchJob).filter(models.BatchJob.id == bjob.id).delete()
    db.flush()



import os
import shutil
from pathlib import Path
from typing import Any, Dict, List, Optional, Union
from uuid import UUID

import jinja2
import yaml
from pydantic import BaseModel, ValidationError

from balsam.client import RequestsClient
from balsam.platform.scheduler import SchedulerInterface

from .config import ClientSettings, InvalidSettings, Settings, SiteConfig, import_string


class SiteDefaults(BaseModel):
    class Config:
        extra = "forbid"

    title: str
    compute_node: str
    mpi_app_launcher: str
    local_app_launcher: str
    mpirun_allows_node_packing: bool
    serial_mode_startup_params: Dict[str, str]
    scheduler_class: str
    allowed_queues: Dict[str, Dict[str, int]]
    allowed_projects: List[str]
    optional_batch_job_params: Dict[str, str]
    globus_endpoint_id: Optional[UUID]


def load_default_configs(top: Optional[Path] = None) -> Dict[Path, SiteDefaults]:
    """
    Load mapping of {default_dir: default_settings_dict} from config/defaults
    """
    if top is None:
        top = Path(__file__).parent.joinpath("defaults")
    default_settings_files = top.glob("*/settings.yml")
    configs_map = {}
    for path in default_settings_files:
        site_dir = path.parent
        with open(path) as fp:
            raw = yaml.safe_load(fp)
            configs_map[site_dir] = SiteDefaults(**raw)
    return configs_map


def render_settings_file(default_settings: Dict[str, Any], template_path: Optional[Path] = None) -> str:
    """
    Generate settings.yml from the master template and a default_settings_dict
    """
    if template_path is None:
        template_path = Path(__file__).parent / "defaults/settings.yml.j2"
    env = jinja2.Environment(trim_blocks=True, lstrip_blocks=True)
    tmpl = env.from_string(template_path.read_text())
    return tmpl.render(default_settings)


def new_site_setup(
    site_path: Union[str, Path],
    name: str,
    default_site_path: Path,
    default_site_conf: SiteDefaults,
    client: Optional["RequestsClient"] = None,
    settings_template_path: Optional[Path] = None,
) -> "SiteConfig":
    """
    Creates a new site directory, registers Site
    with Balsam API, and writes default settings.yml into
    Site directory
    """
    scheduler: SchedulerInterface = import_string(default_site_conf.scheduler_class)
    projects = scheduler.discover_projects()
    default_site_conf.allowed_projects = projects

    if client is None:
        client = ClientSettings.load_from_file().build_client()
    site_path = Path(site_path)
    site_path.mkdir(exist_ok=False, parents=False)

    site_id_file = site_path.joinpath(".balsam-site")

    try:
        site = client.Site.objects.create(name=name, path=site_path)
    except Exception:
        shutil.rmtree(site_path)
        raise
    with open(site_id_file, "w") as fp:
        fp.write(str(site.id))
    os.chmod(site_id_file, 0o440)

    with open(site_path.joinpath("settings.yml"), "w") as fp:
        settings_txt = render_settings_file(default_site_conf.dict(), template_path=settings_template_path)
        fp.write(settings_txt + "\n")
    Path(fp.name).chmod(0o600)

    try:
        settings = Settings.load(fp.name)
    except ValidationError as exc:
        shutil.rmtree(site_path)
        site.delete()
        raise InvalidSettings(f"Invalid Default Settings\n{exc}")

    try:
        cf = SiteConfig(site_path=site_path, settings=settings)
        for path in [cf.log_path, cf.job_path, cf.data_path]:
            path.mkdir(exist_ok=False)
        cf.job_path.chmod(0o700)
        if settings.scheduler is not None:
            job_template_path = settings.scheduler.job_template_path
        else:
            job_template_path = Path("job-template.sh")
        job_template_target = shutil.copy(
            src=default_site_path.joinpath(job_template_path),
            dst=cf.site_path,
        )
        Path(job_template_target).chmod(0o700)

    except FileNotFoundError:
        site.delete()
        shutil.rmtree(site_path)
        raise
    return cf



from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Union

from pydantic import BaseModel, Field, root_validator, validator

from .serializer import serialize

# Set limits to keep queries performant *and* respect the constraints
# of maximum `argv` size that can be passed into a subprocess
# (We can get away with ~1.5MB payload delivered via 128K chunksize args)
MAX_SERIALIZED_PARAMS_SIZE = 512_000


class JobTransferItem(BaseModel):
    location_alias: str
    path: Path

    @validator("path")
    def is_absolute(cls, v: Path) -> Path:
        if not v.is_absolute():
            raise ValueError("Must provide absolute path")
        return v


class JobState(str, Enum):
    created = "CREATED"
    reset = "RESET"
    awaiting_parents = "AWAITING_PARENTS"
    ready = "READY"
    staged_in = "STAGED_IN"
    preprocessed = "PREPROCESSED"
    running = "RUNNING"
    run_done = "RUN_DONE"
    run_error = "RUN_ERROR"
    run_timeout = "RUN_TIMEOUT"
    restart_ready = "RESTART_READY"
    postprocessed = "POSTPROCESSED"
    staged_out = "STAGED_OUT"
    job_finished = "JOB_FINISHED"
    failed = "FAILED"

    @classmethod
    def is_valid(cls, state: str) -> bool:
        return state in cls._value2member_map_


class JobOrdering(str, Enum):
    last_update = "last_update"
    last_update_desc = "-last_update"
    id = "id"
    id_desc = "-id"
    state = "state"
    state_desc = "-state"
    workdir = "workdir"
    workdir_desc = "-workdir"


RUNNABLE_STATES = {JobState.preprocessed, JobState.restart_ready}
DONE_STATES = {JobState.job_finished, JobState.failed}


class JobBase(BaseModel):
    workdir: Path = Field(..., example="test_jobs/test1", description="Job path relative to site data/ folder.")
    tags: Dict[str, str] = Field({}, example={"system": "H2O"}, description="Custom key:value string tags.")
    serialized_parameters: str = Field(
        "", description="Encoded parameters dict", no_constructor=True, no_descriptor=True
    )
    data: Dict[str, Any] = Field({}, example={"energy": -0.5}, description="Arbitrary JSON-able data dictionary.")
    return_code: Optional[int] = Field(None, example=0, description="Return code from last execution of this Job.")

    num_nodes: int = Field(1, example=1, description="Number of compute nodes needed.")
    ranks_per_node: int = Field(1, example=1, description="Number of MPI processes per node.")
    threads_per_rank: int = Field(1, example=1, description="Logical threads per process.")
    threads_per_core: int = Field(1, example=1, description="Logical threads per CPU core.")
    launch_params: Dict[str, str] = Field(
        {},
        example={"cpu_affinity": "depth"},
        description="Optional pass-through parameters to MPI application launcher.",
    )
    gpus_per_rank: float = Field(0, example=0.5, description="Number of GPUs per process.")
    node_packing_count: int = Field(1, example=12, description="Maximum number of concurrent runs per node.")
    wall_time_min: int = Field(
        0,
        example=30,
        description="Optional estimate of Job runtime. All else being equal, longer Jobs tend to run first.",
    )

    @validator("workdir")
    def path_is_relative(cls, v: Path) -> Path:
        if v.is_absolute():
            raise ValueError("Cannot use absolute path")
        return v

    @validator("serialized_parameters")
    def max_params_size(cls, v: str) -> str:
        if len(v) > MAX_SERIALIZED_PARAMS_SIZE:
            raise AssertionError(f"serialized_parameters cannot be larger than {MAX_SERIALIZED_PARAMS_SIZE}")
        return v


class JobCreate(JobBase):
    class Config:
        validate_assignment = True

    # This will generate a Job() constructor signature with `app_id` and `site_name`
    # kwargs. These are used to resolve the integer `app_id` sent to backend.
    app_id: Union[int, str] = Field(..., example=3, description="App name, ID, or class")
    site_name: Optional[str] = Field(
        None,
        example="my-site",
        description="Site name, to disambiguate app defined at multiple Sites.",
        no_export=True,
        no_descriptor=True,
    )

    # Add `parameters` to the constructor, but manage them with a custom getter/setter on
    # JobBase:
    parameters: Dict[str, Any] = Field(
        {},
        example={"name": "world"},
        description="Parameters passed to App at runtime.",
        no_descriptor=True,
        no_export=True,
    )
    parent_ids: Set[int] = Field(set(), example={2, 3}, description="Set of parent Job IDs (dependencies).")
    transfers: Dict[str, JobTransferItem] = Field(
        {},
        example={
            "input_file": {
                "location_alias": "MyCluster",
                "path": "/path/to/input.dat",
            }
        },
        description="TransferItem dictionary. One key:JobTransferItem pair for each slot defined on the App.",
    )

    @root_validator(pre=True)
    def serialize_parameters(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        params = values.get("parameters", {})
        values["serialized_parameters"] = serialize(params)
        return values

    @validator("transfers", pre=True)
    def resolve_transfer_strings(
        cls, transfers: Dict[str, Union[str, JobTransferItem]]
    ) -> Dict[str, JobTransferItem]:
        if not isinstance(transfers, dict):
            raise AssertionError("transfers must be of type dict")
        result = {}
        for k, v in transfers.items():
            if isinstance(v, str):
                loc, path = v.split(":", 1)
                result[k] = JobTransferItem(location_alias=loc, path=path)
            else:
                result[k] = v
        return result

    def dict(self, **kwargs: Any) -> Dict[str, Any]:
        # Overriden to avoid sending "no_export" fields over the wire
        exclude_fields = {f for f, v in JobCreate.__fields__.items() if v.field_info.extra.get("no_export")}
        if "exclude" not in kwargs:
            kwargs["exclude"] = exclude_fields
        elif isinstance(kwargs["exclude"], set):
            kwargs["exclude"].update(exclude_fields)
        else:
            for key in exclude_fields:
                kwargs["exclude"][key] = ...
        return super().dict(**kwargs)


class ServerJobCreate(JobBase):
    app_id: int = Field(..., example=3, description="App ID")
    parent_ids: Set[int] = Field(set(), example={2, 3}, description="Set of parent Job IDs (dependencies).")
    transfers: Dict[str, JobTransferItem] = Field(
        {},
        example={
            "input_file": {
                "location_alias": "MyCluster",
                "path": "/path/to/input.dat",
            }
        },
        description="TransferItem dictionary. One key:JobTransferItem pair for each slot defined on the App.",
    )


class JobUpdate(JobBase):
    workdir: Path = Field(None, example="test_jobs/test1", description="Job path relative to the site data/ folder")
    batch_job_id: int = Field(None, example=4, description="ID of most recent BatchJob in which this Job ran")
    state: JobState = Field(None, example="JOB_FINISHED", description="Job state")
    state_timestamp: datetime = Field(None, description="Time (UTC) at which Job state change occured")
    state_data: Dict[str, Any] = Field({}, description="Arbitrary associated state change data for logging")
    pending_file_cleanup: bool = Field(None, description="Whether job remains to have workdir cleaned.")
    serialized_parameters: str = Field(None, description="Encoded parameters dict", no_descriptor=True)
    serialized_return_value: str = Field(None, description="Encoded return value", no_descriptor=True)
    serialized_exception: str = Field(None, description="Encoded wrapped Exception", no_descriptor=True)

    @validator("serialized_return_value")
    def max_retval_size(cls, v: str) -> str:
        if len(v) > MAX_SERIALIZED_PARAMS_SIZE:
            raise AssertionError(f"serialized_return_value cannot be larger than {MAX_SERIALIZED_PARAMS_SIZE}")
        return v

    @validator("serialized_exception")
    def max_except_size(cls, v: str) -> str:
        if len(v) > MAX_SERIALIZED_PARAMS_SIZE:
            raise AssertionError(f"serialized_except cannot be larger than {MAX_SERIALIZED_PARAMS_SIZE}")
        return v


class JobBulkUpdate(JobUpdate):
    id: int = Field(..., example=123, description="Job id")


class JobOut(JobBase):
    id: int = Field(..., example=22)
    app_id: int = Field(..., example=3)
    parent_ids: Set[int] = Field(..., example={20, 21})
    batch_job_id: int = Field(None, example=4)
    last_update: datetime = Field(...)
    state: JobState = Field(..., example="JOB_FINISHED")
    pending_file_cleanup: bool = Field(..., description="Whether job remains to have workdir cleaned.")
    serialized_parameters: str = Field(..., description="Encoded parameters dict", no_descriptor=True)
    serialized_return_value: str = Field(..., description="Encoded return value", no_descriptor=True)
    serialized_exception: str = Field(..., description="Encoded wrapped Exception", no_descriptor=True)

    class Config:
        orm_mode = True


class PaginatedJobsOut(BaseModel):
    count: int
    results: List[JobOut]



import os
from typing import List, Optional, Union

from .compute_node import ComputeNode


class ThetaKNLNode(ComputeNode):

    cpu_ids = list(range(64))
    gpu_ids: List[Union[int, str]] = []

    @classmethod
    def get_job_nodelist(cls) -> List["ThetaKNLNode"]:
        """
        Get all compute nodes allocated in the current job context
        """
        node_str = os.environ["COBALT_PARTNAME"]
        # string like: 1001-1005,1030,1034-1200
        node_ids = []
        ranges = node_str.split(",")
        lo: Union[str, int]
        hi: Union[int, List[str]]
        for node_range in ranges:
            lo, *hi = node_range.split("-")
            lo = int(lo)
            if hi:
                hi = int(hi[0])
                node_ids.extend(list(range(lo, hi + 1)))
            else:
                node_ids.append(lo)

        return [cls(node_id, f"nid{node_id:05d}") for node_id in node_ids]

    @staticmethod
    def get_scheduler_id() -> Optional[int]:
        id = os.environ.get("COBALT_JOBID")
        if id is not None:
            return int(id)
        return None



import sys

import numpy as np
from numpy.linalg import eigh

# result_file = Path.cwd().with_suffix(".eigs.npy").name  # used for # batch-cluster experiment naming
result_file = "result.npy"  # used by balsam eig.Eig App stage out
input_file = sys.argv[1]
data = np.load(input_file)
print(data.nbytes / 1e9, "GB loaded")
eigvals, eigvecs = eigh(data)
np.save(result_file, eigvals)



import logging
import os
import shutil
import sys
from importlib.util import find_spec
from pathlib import Path
from typing import Dict, Optional, Union

from jinja2 import Template

from balsam.schemas import JobMode

PathLike = Union[str, Path]
logger = logging.getLogger(__name__)


class ScriptTemplate:
    def __init__(self, template_path: PathLike) -> None:
        """
        Wraps a batch job script template located at `template_path`
        """
        with open(template_path) as fp:
            self._template = Template(fp.read())
        logger.debug(f"Loaded job template at {template_path}")

    def render(
        self,
        project: str,
        queue: str,
        num_nodes: int,
        wall_time_min: int,
        job_mode: JobMode,
        filter_tags: Optional[Dict[str, str]] = None,
        partitions: Optional[str] = None,
        **kwargs: str,
    ) -> str:
        """
        Returns string contained filled batch job script
        """
        conf = dict(
            project=project,
            queue=queue,
            num_nodes=num_nodes,
            wall_time_min=wall_time_min,
            job_mode=job_mode,
            filter_tags=filter_tags,
            partitions=partitions,
            optional_params=kwargs,
        )
        conf["balsam_site_path"] = os.environ["BALSAM_SITE_PATH"]
        conf["balsam_bin"] = self.locate_balsam()
        conf["pg_bin"] = self.locate_postgres()
        conf["launcher_cmd"] = self.launcher_entrypoint()
        return self._template.render(conf)

    @staticmethod
    def locate_balsam() -> Path:
        balsam_bin = shutil.which("balsam") or sys.executable
        return Path(balsam_bin).parent

    @staticmethod
    def locate_postgres() -> Optional[Path]:
        pg_bin = shutil.which("pg_ctl") or os.environ.get("POSTGRES_BIN")
        return Path(pg_bin).parent if pg_bin else None

    @staticmethod
    def launcher_entrypoint() -> str:
        spec = find_spec("balsam.cmdline.launcher")
        if spec is None:
            raise RuntimeError("Could not locate balsam.cmdline.launcher module")
        launcher_path = spec.origin
        if launcher_path is None:
            raise RuntimeError("Could not locate balsam.cmdline.launcher module")
        path = Path(launcher_path).resolve()
        return f"{sys.executable} {path}"



import logging

from alembic import context  # type: ignore

from balsam.server import settings
from balsam.server.models import Base, get_engine

logger = logging.getLogger("balsam.server.models.alembic.env")


metadata = Base.metadata
engine = get_engine()
logger.info(f"Alembic running migrations with DB engine: {engine}")

with engine.connect() as conn:
    context.configure(connection=conn, target_metadata=metadata, include_schemas=True)
    with context.begin_transaction():
        context.run_migrations()



from dateutil.parser import isoparse
from fastapi import status

from balsam.server import models

from .util import create_site


def test_create_site(auth_client):
    posted_site = create_site(
        auth_client,
        name="thetalogin3.alcf.anl.gov",
        path="/projects/myProject/balsam-site",
    )
    assert type(posted_site["id"]) == int
    site_list = auth_client.get("/sites/")["results"]
    assert isinstance(site_list, list)
    assert len(site_list) == 1
    assert site_list[0] == posted_site


def test_cannot_create_duplicate_site(auth_client):
    create_site(
        auth_client,
        name="theta",
        path="/projects/mysite1",
        check=status.HTTP_201_CREATED,
    )
    create_site(
        auth_client,
        name="theta",
        path="/projects/mysite1",
        check=status.HTTP_400_BAD_REQUEST,
    )


def test_detail_view(auth_client):
    created_site = create_site(auth_client)
    id = created_site["id"]
    retrieved_site = auth_client.get(f"sites/{id}")
    assert retrieved_site == created_site


def test_update_site_status(auth_client):
    created_site = create_site(auth_client)
    id = created_site["id"]
    created_time = isoparse(created_site["last_refresh"])
    created_site.update(dict(num_nodes=128))
    updated_site = auth_client.put(f"sites/{id}", **created_site)
    updated_time = isoparse(updated_site["last_refresh"])
    assert updated_time > created_time


def test_cannot_partial_update_owner(auth_client, db_session):
    created_site = create_site(auth_client)
    id = created_site["id"]
    owner_id = db_session.query(models.Site.owner_id).one()[0]
    other_owner_id = owner_id + 10
    auth_client.put(f"sites/{id}", owner_id=other_owner_id)
    assert db_session.query(models.Site.owner_id).one()[0] == owner_id


def test_deleting_site_removes_associated_apps(auth_client):
    site = create_site(auth_client)
    auth_client.post(
        "/apps/",
        site_id=site["id"],
        name="HelloWorld",
        serialized_class="txt",
        source_code="txt",
        parameters={"name": {"required": True}, "N": {"required": False, "default": 1}},
    )
    assert len(auth_client.get("/apps/")["results"]) == 1
    auth_client.delete(f"/sites/{site['id']}", check=status.HTTP_204_NO_CONTENT)
    assert len(auth_client.get("/apps/")["results"]) == 0



import random
import time
from pathlib import Path
from typing import Any, Dict, List

import click

from balsam.api import App, Job, Site

APP_CLASS = "eig.Eig"
SITE_ID = 11
REMOTE_ALIAS = "aps_dtn"
RESULT_DIR: Path = Path("/gdata/lcfwork/results-2021-03-24/")
SITE = Site.objects.get(id=SITE_ID)
assert SITE.id is not None
APP = App.objects.get(site_id=SITE.id, class_path=APP_CLASS)

SIZE_TO_FILES = {
    "S": ["/gdata/lcfwork/eig-inputs/5000.npy"],
    "L": ["/gdata/lcfwork/eig-inputs/12_000.npy"],
    "C": ["/gdata/lcfwork/eig-inputs/5000.npy", "/gdata/lcfwork/eig-inputs/12_000.npy"],
}


def populate_jobs(num_nodes: int, size_tag: str, num_jobs: int, exp_tag: str) -> None:
    tags = {
        "experiment": exp_tag,
        "job_source": REMOTE_ALIAS,
        "num_nodes": str(num_nodes),
        "size": size_tag,
    }
    jobs: List[Job] = []
    for i in range(num_jobs):
        workdir = Path(exp_tag) / (str(num_nodes) + str(size_tag)) / f"{i:06d}"
        inp_path = Path(random.choice(SIZE_TO_FILES[size_tag]))
        assert inp_path.is_absolute()
        result_path = RESULT_DIR.joinpath(inp_path.name).with_suffix(f".eig{i:06d}.npy")
        transfers: Dict[str, Any] = {
            "matrix": {
                "location_alias": REMOTE_ALIAS,
                "path": inp_path,
            },
            "eigvals": {
                "location_alias": REMOTE_ALIAS,
                "path": result_path,
            },
        }
        assert APP.id is not None
        job = Job(
            app_id=APP.id,
            workdir=workdir,
            tags=tags,
            parameters={"inp_file": inp_path.name},
            transfers=transfers,
            threads_per_rank=64,
            launch_params={"cpu_affinity": "depth"},
        )
        jobs.append(job)
    Job.objects.bulk_create(jobs)


@click.command()
@click.option("-t", "--time-limit-min", type=int, required=True)
@click.option("-n", "--num-nodes", type=int, required=True)
@click.option("-s", "--size", type=click.Choice(["L", "C"]), required=True)
@click.option("-e", "--experiment", required=True)
def main(time_limit_min: int, num_nodes: int, size: str, experiment: str) -> None:
    start = time.time()
    time_limit_sec = time_limit_min * 60
    tags = {
        "experiment": experiment,
        "job_source": REMOTE_ALIAS,
        "num_nodes": str(num_nodes),
        "size": size,
    }
    while (time.time() - start) < time_limit_sec:
        qs = Job.objects.filter(app_id=APP.id, tags=tags, state="READY")  # type: ignore
        to_create = max(0, 48 - qs.count())  # type: ignore
        if to_create:
            populate_jobs(num_nodes=num_nodes, size_tag=size, num_jobs=to_create, exp_tag=experiment)
            print(f"Added {to_create} jobs to pipeline: {tags}")
        else:
            print("Pipeline already full; not adding jobs")
        time.sleep(2 + random.random())
    print("Done!")


if __name__ == "__main__":
    main()



import logging
from typing import Iterator

from sqlalchemy import create_engine, orm
from sqlalchemy.engine import Engine
from sqlalchemy.ext.declarative import declarative_base

import balsam.server

logger = logging.getLogger(__name__)

Base = declarative_base()
_engine = None
_Session = None


def get_engine() -> Engine:
    global _engine
    if _engine is None:
        logger.info(f"Creating DB engine: {balsam.server.settings.database_url}")
        _engine = create_engine(
            balsam.server.settings.database_url,
            connect_args={},
            pool_size=10,
            max_overflow=40,
        )
    return _engine


def get_session() -> Iterator[orm.Session]:
    global _Session
    if _Session is None:
        _Session = orm.sessionmaker(bind=get_engine())

    session: orm.Session = _Session()
    try:
        yield session
    except:  # noqa: E722
        session.rollback()
        raise
    finally:
        session.close()


def create_tables() -> None:
    Base.metadata.create_all(get_engine())



from typing import Any, Dict, Optional, Tuple, cast

from fastapi.encoders import jsonable_encoder
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import Query, Session

from balsam import schemas
from balsam.server import ValidationError, models
from balsam.server.routers.filters import AppQuery
from balsam.server.utils import Paginator


def fetch(
    db: Session,
    owner: schemas.UserOut,
    paginator: Optional[Paginator[models.App]] = None,
    app_id: Optional[int] = None,
    filterset: Optional[AppQuery] = None,
) -> "Tuple[int, Query[models.App]]":
    qs = db.query(models.App).join(models.Site).filter(models.Site.owner_id == owner.id)  # type: ignore
    if filterset is not None:
        qs = filterset.apply_filters(qs)
    if app_id is not None:
        qs = qs.filter(models.App.id == app_id).one()
        return (1, qs)
    else:
        assert paginator is not None
        count = qs.group_by(models.App.id).count()
        return count, paginator.paginate(qs)


def flush_or_400(db: Session) -> None:
    try:
        db.flush()
    except IntegrityError as e:
        if "duplicate key value" in str(e):
            db.rollback()
            raise ValidationError("An app with this site_id and name already exists")
        else:
            raise


def create(db: Session, owner: schemas.UserOut, app: schemas.AppCreate) -> models.App:
    # Will raise if the owner does not have a matching Site ID
    app.site_id = db.query(models.Site.id).filter_by(id=app.site_id, owner_id=owner.id).one()[0]
    new_app = models.App(**jsonable_encoder(app))
    db.add(new_app)
    flush_or_400(db)
    return new_app


def update(db: Session, owner: schemas.UserOut, app_id: int, update_data: Dict[str, Any]) -> models.App:
    qs = (
        db.query(models.App)
        .join(models.Site, models.App.site_id == models.Site.id)  # type: ignore
        .filter(models.Site.owner_id == owner.id, models.App.id == app_id)
    )
    app_in_db = cast(models.App, qs.one())
    if "site_id" in update_data:
        update_data["site_id"] = (
            db.query(models.Site.id).filter_by(id=update_data["site_id"], owner_id=owner.id).one()[0]
        )
    for k, v in update_data.items():
        setattr(app_in_db, k, v)
    flush_or_400(db)
    return app_in_db


def delete(db: Session, owner: schemas.UserOut, app_id: int) -> None:
    app_in_db = (
        db.query(models.App).join(models.Site).filter(models.Site.owner_id == owner.id, models.App.id == app_id).one()  # type: ignore
    )
    db.delete(app_in_db)  # type: ignore
    db.flush()



from typing import Any, Dict, List, Optional, Type, TypeVar, Union

IntStr = Union[int, str]

U = TypeVar("U")


class ComputeNode:

    cpu_ids: List[IntStr] = []
    gpu_ids: List[IntStr] = []

    def __init__(self, node_id: IntStr, hostname: str, gpu_ids: Optional[List[IntStr]] = None) -> None:
        self.node_id = node_id
        self.hostname = hostname
        self.occupancy = 0.0
        self.jobs: Dict[int, Dict[str, Any]] = {}
        self.idle_cpus: List[IntStr] = [i for i in self.cpu_ids]
        self.busy_cpus: List[IntStr] = []
        if gpu_ids is None:
            gpu_ids = self.gpu_ids
        self.idle_gpus: List[IntStr] = [i for i in gpu_ids]
        self.busy_gpus: List[IntStr] = []

    def check_fit(self, num_cpus: int, num_gpus: int, occupancy: float) -> bool:
        if self.occupancy + occupancy > 1.001:
            return False
        elif num_cpus > len(self.idle_cpus):
            return False
        elif num_gpus > len(self.idle_gpus):
            return False
        else:
            return True

    def assign(self, job_id: int, num_cpus: int = 0, num_gpus: int = 0, occupancy: float = 1.0) -> Dict[str, Any]:
        if job_id in self.jobs:
            raise ValueError(f"Already have job {job_id}")

        self.occupancy += occupancy
        if self.occupancy > 0.999:
            self.occupancy = 1.0

        assigned_cpus = self.idle_cpus[:num_cpus]
        assigned_gpus = self.idle_gpus[:num_gpus]
        if assigned_cpus:
            self.busy_cpus.extend(assigned_cpus)
            self.idle_cpus = [i for i in self.idle_cpus if i not in assigned_cpus]
        if assigned_gpus:
            self.busy_gpus.extend(assigned_gpus)
            self.idle_gpus = [i for i in self.idle_gpus if i not in assigned_gpus]
        resource_spec = {
            "cpu_ids": assigned_cpus,
            "gpu_ids": assigned_gpus,
            "occupancy": occupancy,
        }
        self.jobs[job_id] = resource_spec
        return resource_spec

    def free(self, job_id: int) -> None:
        resource_spec = self.jobs.pop(job_id)
        self.occupancy -= resource_spec["occupancy"]
        if self.occupancy < 0.001:
            self.occupancy = 0.0
        cpu_ids = resource_spec["cpu_ids"]
        gpu_ids = resource_spec["gpu_ids"]
        if cpu_ids:
            self.idle_cpus.extend(cpu_ids)
            self.busy_cpus = [i for i in self.busy_cpus if i not in cpu_ids]
        if gpu_ids:
            self.idle_gpus.extend(gpu_ids)
            self.busy_gpus = [i for i in self.busy_gpus if i not in gpu_ids]

    @classmethod
    def get_job_nodelist(cls: Type[U]) -> "List[U]":
        """
        Get all compute nodes allocated in the current job context
        """
        return []

    @staticmethod
    def get_scheduler_id() -> Optional[int]:
        return None

    def __repr__(self) -> str:
        busy_cpus = len(self.busy_cpus)
        total_cpus = len(self.cpu_ids)
        busy_gpus = len(self.busy_gpus)
        total_gpus = len(self.gpu_ids)
        cpu_str = f"{busy_cpus}/{total_cpus} CPUs busy"
        gpu_str = f", {busy_gpus}/{total_gpus} GPUs busy" if total_gpus else ""
        d = dict(
            node_id=self.node_id,
            hostname=self.hostname,
            occupancy=self.occupancy,
            num_jobs=len(self.jobs),
        )
        args = ", ".join(f"{k}={v}" for k, v in d.items())
        rep = f"{self.__class__.__name__}({args}, {cpu_str}{gpu_str})"
        return rep



import logging
import os
from pathlib import Path
from typing import List, Optional, Union

from .compute_node import ComputeNode

logger = logging.getLogger(__name__)
IntStr = Union[int, str]


class ThetaGPUNode(ComputeNode):

    cpu_ids = list(range(128))
    gpu_ids: List[IntStr] = list(range(8))

    @classmethod
    def get_job_nodelist(cls) -> List["ThetaGPUNode"]:
        """
        Get all compute nodes allocated in the current job context
        """
        nodefile = os.environ["COBALT_NODEFILE"]
        # a file containing a list of node hostnames, one per line
        # thetagpu01
        # thetagpu02
        with open(nodefile) as fp:
            data = fp.read()
        splitter = "," if "," in data else None
        hostnames = data.split(splitter)
        hostnames = [h.strip() for h in hostnames if h.strip()]
        node_ids = [int(hostname[-2:]) for hostname in hostnames]
        node_list = []
        for nid, hostname in zip(node_ids, hostnames):
            gpu_ids = cls.discover_gpu_list(hostname)
            node_list.append(cls(nid, hostname, gpu_ids=gpu_ids))
        return node_list

    @classmethod
    def discover_gpu_list(cls, hostname: str) -> List[IntStr]:
        gpu_file = Path(f"/var/tmp/balsam-{hostname}-gpulist.txt")
        gpu_ids: List[IntStr]
        if gpu_file.is_file():
            tokens = gpu_file.read_text().split()
            gpu_ids = [t[:-1] for t in tokens if t.startswith("MIG-GPU-")]
        else:
            gpu_ids = cls.gpu_ids
        logger.info(f"{hostname} detected GPU IDs: {gpu_ids}")
        return gpu_ids

    @staticmethod
    def get_scheduler_id() -> Optional[int]:
        id = os.environ.get("COBALT_JOBID")
        if id is not None:
            return int(id)
        return None



import uuid
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List

from pydantic import BaseModel, Field


class TransferItemState(str, Enum):
    awaiting_job = "awaiting_job"
    pending = "pending"
    active = "active"
    inactive = "inactive"
    done = "done"
    error = "error"


class TransferProtocol(str, Enum):
    globus = "globus"
    rsync = "rsync"


class TransferDirection(str, Enum):
    stage_in = "in"
    stage_out = "out"


class TransferItemBase(BaseModel):
    state: TransferItemState = Field(..., description="Status of this transfer item", example="active")
    task_id: str = Field("", example=uuid.uuid4(), description="Transfer Task ID used to lookup transfer item status")
    transfer_info: Dict[str, Any] = Field(
        {}, example={"bandwidth": "4028"}, description="Arbitrary transfer state info"
    )


class TransferItemOut(TransferItemBase):
    id: int = Field(..., description="Transfer item id", example=2)
    job_id: int = Field(..., description="Associated Job id", example=123)
    direction: TransferDirection = Field(..., description="Transfer direction (in/out)", example="in")
    local_path: Path = Field(..., description="Path relative to the Job workdir", example="input.json")
    remote_path: Path = Field(
        ..., description="Absolute path on the remote filesystem", example="/path/to/input.json"
    )
    location_alias: str = Field(..., description="Site-defined alias of the remote location", example="APS-DTN")
    recursive: bool = Field(..., description="Whether or not this item is a directory", example=False)

    class Config:
        orm_mode = True


class PaginatedTransferItemOut(BaseModel):
    count: int
    results: List[TransferItemOut]


class TransferItemUpdate(TransferItemBase):
    state: TransferItemState = Field(None, example="active", description="Status of this transfer item")


class TransferItemBulkUpdate(TransferItemUpdate):
    id: int



import json
import os
import signal
import subprocess
import tempfile
from pathlib import Path
from typing import Union

import click
import jinja2

import balsam.server

REDIS_TMPL = Path(balsam.server.__file__).parent.joinpath("redis.conf.tmpl")

# NOTE: lazy-import balsam.util.postgres inside each CLI handler
# Because psycopg2 is slow to import


@click.group()
def server() -> None:
    """
    Deploy and manage a local Balsam server
    """
    pass


@server.command()
@click.option("-p", "--path", required=True, type=click.Path(exists=True, file_okay=False))
def down(path: Union[str, Path]) -> None:
    """
    Shut down the Balsam server (Postgres, Redis, and Gunicorn)
    """
    path = Path(path).resolve()
    pid_files = ["balsamdb/postmaster.pid", "redis.pid", "gunicorn.pid"]
    services = ["PostgreSQL", "Redis", "Gunicorn"]
    for serv, pid_file in zip(services, pid_files):
        try:
            with open(path.joinpath(pid_file)) as fp:
                pid = int(fp.readline())
        except (FileNotFoundError, ValueError):
            click.echo(f"Could not read pidfile for stopping {serv}")
        else:
            click.echo(f"Sending SIGTERM to {serv} (pid={pid})")
            try:
                os.kill(pid, signal.SIGTERM)
            except ProcessLookupError:
                click.echo(f"No process found with pid={pid}")


@server.command()
@click.option("-p", "--path", required=True, type=click.Path(exists=True, file_okay=False))
@click.option("--log/--no-log", is_flag=True, default=False)
def up(path: Union[Path, str], log: bool) -> None:
    """
    Starts up the services comprising the Balsam server (Postgres, Redis, and Gunicorn)
    """
    from balsam.util import postgres as pg

    path = Path(path).resolve()
    gunicorn_config_file = Path(balsam.server.__file__).parent / "gunicorn.conf.example.py"
    assert gunicorn_config_file.is_file()

    click.echo("Starting Redis daemon")
    try:
        start_redis(path)
    except RuntimeError:
        click.echo("Skipping Redis")
    db_path = path.joinpath("balsamdb").as_posix()
    pg.start_db(db_path)
    dsn = pg.load_dsn(db_path)
    pg.configure_balsam_server_from_dsn(dsn)

    os.environ["BALSAM_LOG_DIR"] = path.as_posix() if log else ""
    args = [
        "gunicorn",
        "-c",
        gunicorn_config_file.as_posix(),
        "balsam.server.main:app",
    ]
    p = subprocess.Popen(args, cwd=path)
    click.echo(f"Started gunicorn (pid={p.pid})")


@server.command()
@click.option("-p", "--path", required=True, type=click.Path(exists=False, writable=True))
def deploy(path: Union[Path, str]) -> None:
    """
    Create a new Balsam database and API server instance
    """
    path = Path(path).resolve()
    if path.exists():
        raise click.BadParameter(f"{path} already exists")

    from balsam.util import postgres as pg

    gunicorn_config_file = Path(balsam.server.__file__).parent / "gunicorn.conf.example.py"
    assert gunicorn_config_file.is_file()

    click.echo("Creating database")
    pw_dict = pg.create_new_db(db_path=path.joinpath("balsamdb"))
    click.echo("Database created!")

    # Point balsam.server.settings at new DSN:
    dsn = pg.configure_balsam_server(**pw_dict)

    click.echo("Running alembic migrations")
    pg.run_alembic_migrations(dsn)
    click.echo("Migrations complete!")

    write_redis_conf(path.joinpath("redis.conf"))
    click.echo("Starting Redis daemon")
    try:
        start_redis(path)
    except RuntimeError:
        click.echo("Skipping Redis")

    os.environ["BALSAM_LOG_DIR"] = path.as_posix()
    args = [
        "gunicorn",
        "-c",
        gunicorn_config_file.as_posix(),
        "balsam.server.main:app",
    ]
    p = subprocess.Popen(args, cwd=path)
    click.echo(f"Started gunicorn (pid={p.pid})")


@server.command()
def migrate() -> None:
    from balsam.util import postgres as pg

    dsn = balsam.server.Settings().database_url
    click.echo(f"Running alembic migrations for {dsn}")
    pg.run_alembic_migrations(dsn)
    click.echo("Migrations complete!")


def write_redis_conf(conf_path: Path) -> None:
    tmpl = jinja2.Template(REDIS_TMPL.read_text())
    with tempfile.NamedTemporaryFile(prefix="redis-balsam", suffix=".sock") as fp:
        tmp_name = Path(fp.name).resolve().as_posix()
    conf = tmpl.render({"unix_sock_path": tmp_name, "run_path": str(conf_path.parent)})
    with open(conf_path, "w") as f:
        f.write(conf)
    click.echo(f"Wrote redis conf to: {conf_path}")


def start_redis(path: Path, config_filename: str = "redis.conf") -> None:
    config_file = path.joinpath(config_filename)
    with open(config_file) as fp:
        for line in fp:
            if line.strip().startswith("unixsocket"):
                sock_file = line.split()[1]
                os.environ["balsam_redis_params"] = json.dumps({"unix_socket_path": sock_file})
    proc = subprocess.run(
        f"redis-server {config_file} --daemonize yes",
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        encoding="utf-8",
        shell=True,
        check=False,
        cwd=path,
    )
    if proc.returncode != 0:
        click.echo(f"Error in starting Redis:\n{proc.stdout}")
        raise RuntimeError
    click.echo("Started redis daemon")



from typing import Any, Dict, List

from fastapi import APIRouter, Depends, status
from sqlalchemy import orm

from balsam import schemas
from balsam.server import settings
from balsam.server.models import BatchJob, crud, get_session
from balsam.server.pubsub import pubsub
from balsam.server.utils import Paginator

from .filters import BatchJobQuery

router = APIRouter()
auth = settings.auth.get_auth_method()


@router.get("/", response_model=schemas.PaginatedBatchJobOut)
def list(
    db: orm.Session = Depends(get_session),
    user: schemas.UserOut = Depends(auth),
    paginator: Paginator[BatchJob] = Depends(Paginator),
    q: BatchJobQuery = Depends(BatchJobQuery),
) -> Dict[str, Any]:
    """List the BatchJobs submitted at the user's Balsam Sites."""
    count, batch_jobs = crud.batch_jobs.fetch(db, owner=user, paginator=paginator, filterset=q)
    return {"count": count, "results": batch_jobs}


@router.get("/{batch_job_id}", response_model=schemas.BatchJobOut)
def read(
    batch_job_id: int, db: orm.Session = Depends(get_session), user: schemas.UserOut = Depends(auth)
) -> BatchJob:
    """Get a BatchJob by id."""
    count, batch_jobs = crud.batch_jobs.fetch(db, owner=user, batch_job_id=batch_job_id)
    return batch_jobs[0]


@router.post("/", response_model=schemas.BatchJobOut, status_code=status.HTTP_201_CREATED)
def create(
    batch_job: schemas.BatchJobCreate, db: orm.Session = Depends(get_session), user: schemas.UserOut = Depends(auth)
) -> schemas.BatchJobOut:
    """Submit a new BatchJob to a Balsam Site's queue."""
    new_batch_job = crud.batch_jobs.create(db, owner=user, batch_job=batch_job)
    result = schemas.BatchJobOut.from_orm(new_batch_job)
    db.commit()
    pubsub.publish(user.id, "create", "batch_job", result)
    return result


@router.put("/{batch_job_id}", response_model=schemas.BatchJobOut)
def update(
    batch_job_id: int,
    batch_job: schemas.BatchJobUpdate,
    db: orm.Session = Depends(get_session),
    user: schemas.UserOut = Depends(auth),
) -> schemas.BatchJobOut:
    """Update a BatchJob by id."""
    updated_batch_job = crud.batch_jobs.update(db, owner=user, batch_job_id=batch_job_id, batch_job=batch_job)
    result = schemas.BatchJobOut.from_orm(updated_batch_job)
    db.commit()
    pubsub.publish(user.id, "update", "batch_job", result)
    return result


@router.patch("/", response_model=List[schemas.BatchJobOut])
def bulk_update(
    batch_jobs: List[schemas.BatchJobBulkUpdate],
    db: orm.Session = Depends(get_session),
    user: schemas.UserOut = Depends(auth),
) -> List[schemas.BatchJobOut]:
    """Update a list of BatchJobs."""
    updated_batch_jobs = crud.batch_jobs.bulk_update(db, owner=user, batch_jobs=batch_jobs)
    result = [schemas.BatchJobOut.from_orm(j) for j in updated_batch_jobs]
    db.commit()
    pubsub.publish(user.id, "bulk-update", "batch_job", result)
    return result


@router.delete("/{batch_job_id}", status_code=status.HTTP_204_NO_CONTENT)
def delete(batch_job_id: int, db: orm.Session = Depends(get_session), user: schemas.UserOut = Depends(auth)) -> None:
    """Delete a BatchJob by id."""
    crud.batch_jobs.delete(db, owner=user, batch_job_id=batch_job_id)
    db.commit()
    pubsub.publish(user.id, "delete", "batch_job", {"id": batch_job_id})



import os
import socket
import subprocess
import sys
import time
from datetime import datetime, timedelta
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, TypeVar, Union, cast

import click
import psutil  # type: ignore

from balsam._api.models import AppQuery, BatchJobQuery, JobQuery, Site
from balsam.config import ClientSettings, SiteConfig
from balsam.schemas import BatchJobPartition

if TYPE_CHECKING:
    from balsam.client import RESTClient


AppJobQuery = Union[AppQuery, JobQuery, BatchJobQuery]
T = TypeVar("T", AppQuery, JobQuery, BatchJobQuery)

PID_FILENAME = "balsam-service.pid"


def utc_past(minutes_ago: int) -> datetime:
    return datetime.utcnow() - timedelta(minutes=minutes_ago)


def is_site_active(s: Site, threshold_min: int = 2) -> bool:
    threshold = utc_past(minutes_ago=threshold_min)
    if s.last_refresh is None:
        return False
    return s.last_refresh >= threshold


def filter_by_sites(query: T, site_str: str = "") -> T:
    """
    Applies the appropriate site_id filter to a query based on site selector string.
    - --site=all does not filter by site_id
    - --site=this explicitly filters by local site_id
    - --site=active filters by active sites only
    - otherwise, a comma-separated list of ID's or Path fragments
    Default behavior (no --site argument):
        - Select the local site_id if available
        - Otherwise, select all active sites
    """
    Site = query._manager._client.Site
    try:
        site_conf: Optional[SiteConfig] = SiteConfig()
    except ValueError:
        site_conf = None
    site_id: Optional[int] = site_conf.site_id if site_conf else None

    values = [s.strip() for s in site_str.split(",") if s.strip()]
    active_t = utc_past(minutes_ago=2)
    if not values:
        if site_id is not None:
            return query.filter(site_id=site_id)
        else:
            active_ids = [site.id for site in Site.objects.filter(last_refresh_after=active_t) if site.id]
            return query.filter(site_id=active_ids)

    if "all" in values:
        return query
    if "active" in values:
        active_ids = [site.id for site in Site.objects.filter(last_refresh_after=active_t) if site.id]
        return query.filter(site_id=active_ids)
    if "this" in values:
        if site_id is None:
            raise click.BadParameter(
                "Cannot use --site=this outside of a Balsam Site. "
                "Please navigate into a Balsam site directory, or set "
                "BALSAM_SITE_PATH"
            )
        return query.filter(site_id=site_id)

    site_ids, path_fragments = [], []
    for v in values:
        if v.isdigit():
            site_ids.append(int(v))
        else:
            path_fragments.append(v)

    for path_fragment in path_fragments:
        site_ids.extend(site.id for site in Site.objects.filter(path=path_fragment) if site.id)

    return query.filter(site_id=site_ids)


def list_to_dict(arg_list: List[str]) -> Dict[str, str]:
    return dict(cast(Tuple[str, str], arg.split("=", maxsplit=1)) for arg in arg_list)


def validate_tags(ctx: Any, param: Any, value: List[str]) -> Dict[str, str]:
    try:
        return list_to_dict(value)
    except ValueError:
        raise click.BadParameter("needs to be in format KEY=VALUE")


def validate_partitions(ctx: Any, param: Any, value: List[str]) -> List[BatchJobPartition]:
    partitions = []
    for arg in value:
        try:
            job_mode, num_nodes, *filter_tags_list = arg.split(":")
        except ValueError:
            raise click.BadParameter("needs to be in format MODE:NUM_NODES[:KEY=VALUE]")
        filter_tags: Dict[str, str] = validate_tags(ctx, param, filter_tags_list)
        partitions.append(BatchJobPartition(job_mode=job_mode, num_nodes=num_nodes, filter_tags=filter_tags))
    return partitions


def load_site_config() -> SiteConfig:
    try:
        cf = SiteConfig()
    except ValueError:
        raise click.BadParameter(
            "Cannot perform this action outside the scope of a Balsam Site. "
            "Please navigate into a Balsam site directory, or set "
            "BALSAM_SITE_PATH"
        )
    return cf


def load_site_from_selector(site_selector: str) -> Site:
    client = load_client()
    Site = client.Site
    if not site_selector:
        site_config: SiteConfig = load_site_config()
        site = Site.objects.get(id=site_config.site_id)
    elif site_selector.isdigit():
        site = Site.objects.get(id=int(site_selector))
    else:
        site = Site.objects.get(path=site_selector)
    return site


def load_client() -> "RESTClient":
    return ClientSettings.load_from_file().build_client()


def get_pidfile(site_config: SiteConfig) -> Path:
    site_dir = site_config.site_path
    pid_file: Path = site_dir.joinpath(PID_FILENAME)
    return pid_file


def read_pidfile(site_config: SiteConfig) -> Tuple[str, int]:
    site_dir = site_config.site_path
    pid_file: Path = site_dir.joinpath(PID_FILENAME)
    service_host, service_pid = pid_file.read_text().split("\n")[:2]
    return service_host, int(service_pid)


def start_site(site_dir: Path) -> "subprocess.Popen[bytes]":
    os.environ["BALSAM_SITE_PATH"] = site_dir.as_posix()
    p = subprocess.Popen(
        [sys.executable, "-m", "balsam.site.service.main"],
        cwd=site_dir,
    )
    time.sleep(0.2)
    if p.poll() is None:
        return p
    raise RuntimeError(f"balsam.site.service.main return code {p.poll()}")


def check_killable(cf: SiteConfig, raise_exc: bool = False) -> Optional[int]:
    pid_file = get_pidfile(cf)
    cur_host = socket.gethostname()

    if not pid_file.is_file():
        if raise_exc:
            raise click.BadArgumentUsage(f"There is no {PID_FILENAME} in {cf.site_path}")
        return None
    try:
        service_host, service_pid = read_pidfile(cf)
    except ValueError:
        if raise_exc:
            raise click.BadArgumentUsage(
                f"Failed to read {pid_file}: please delete it and manually kill the site daemon process!"
            )
        return None
    if cur_host != service_host:
        if raise_exc:
            raise click.BadArgumentUsage(
                f"The site daemon is running on {service_host}; cannot stop from current host: {cur_host}"
            )
        return None
    if not psutil.pid_exists(service_pid):
        if raise_exc:
            raise click.BadArgumentUsage(
                f"Could not find process with PID {service_pid} on {service_host}. "
                f"Make sure the Balsam site daemon isn't running and delete {pid_file}"
            )
        return None
    return service_pid


def kill_site(cf: SiteConfig, service_pid: int) -> None:
    service_host = socket.gethostname()
    try:
        service_proc = psutil.Process(pid=service_pid)
        service_proc.terminate()
    except (ProcessLookupError, psutil.ProcessLookupError):
        raise click.BadArgumentUsage(
            f"Could not find process with PID {service_pid} on {service_host}. "
            f"Make sure the Balsam site daemon isn't running and delete {get_pidfile(cf)}"
        )
    click.echo(f"Sent SIGTERM to Balsam site daemon [pid {service_pid}]")
    click.echo(f"Waiting for site {cf.site_path} to shutdown...")
    with click.progressbar(range(12)) as bar:
        for i in bar:
            try:
                service_proc.wait(timeout=1)
            except psutil.TimeoutExpired:
                if i == 11:
                    raise click.BadArgumentUsage(
                        f"Site daemon did not shut down gracefully on its own; please kill it manually "
                        f"and delete {get_pidfile(cf)}"
                    )
            else:
                click.echo("\nSite daemon shutdown OK")
                break


def table_print(data: List[Dict[str, Any]]) -> None:
    if not data:
        return
    col_names = list(data[0].keys())
    max_widths = [len(header) + 2 for header in col_names]
    rows = []
    for record in data:
        row = [str(record[col_name]) for col_name in col_names]
        for i, entry in enumerate(row):
            max_widths[i] = max(max_widths[i], len(entry) + 2)
        rows.append(row)

    header = " ".join(col_name.ljust(width) for col_name, width in zip(col_names, max_widths))
    click.echo(header)

    for row in rows:
        line = " ".join(row[i].ljust(width) for i, width in enumerate(max_widths))
        click.echo(line)



from .alcf_cooley_node import CooleyNode
from .alcf_thetagpu_node import ThetaGPUNode
from .alcf_thetaknl_node import ThetaKNLNode
from .compute_node import ComputeNode
from .default import DefaultNode
from .nersc_corihas_node import CoriHaswellNode
from .nersc_coriknl_node import CoriKNLNode
from .summit_node import SummitNode

__all__ = [
    "DefaultNode",
    "ThetaKNLNode",
    "SummitNode",
    "ThetaGPUNode",
    "CooleyNode",
    "CoriHaswellNode",
    "CoriKNLNode",
    "ComputeNode",
]



import random
import time
from datetime import datetime, timedelta
from pathlib import Path
from typing import Any, Dict, List

import yaml
from pydantic import BaseSettings

from balsam.api import App, BatchJob, Job, Site
from balsam.schemas import JobState

APP_CLASS = "eig.Eig"
RESULT_IDX = 0
ACTIVE_STATES = set(
    [
        JobState.created,
        JobState.ready,
        JobState.awaiting_parents,
        JobState.staged_in,
        JobState.preprocessed,
        JobState.running,
    ]
)


class RunConf(BaseSettings):
    time_limit_min: int = 16
    remote_alias: str = "aps_dtn"
    result_dir: Path = Path("/gdata/lcfwork/results-2021-03-24/")
    num_nodes: int
    jobs_per_sec: float
    matrix_inputs: List[Path]
    size_tag: str


class Conf(BaseSettings):
    project: str = "WorkExpFacil"
    queue: str = "R.WorkExpFacil"
    site_id: int = 11
    experiment_tag: str
    run_list: List[RunConf]


def submit_eig_job(
    app_id: int, workdir: Path, tags: Dict[str, str], remote_alias: str, matrix_inp_path: Path, result_dir: Path
) -> Job:
    result_path = result_dir.joinpath(matrix_inp_path.name).with_suffix(f".eig{RESULT_IDX:06d}.npy")
    transfers: Dict[str, Any] = {
        "matrix": {
            "location_alias": remote_alias,
            "path": matrix_inp_path,
        },
        "eigvals": {
            "location_alias": remote_alias,
            "path": result_path,
        },
    }
    j = Job(
        app_id=app_id,
        workdir=workdir,
        tags=tags,
        parameters={"inp_file": matrix_inp_path.name},
        transfers=transfers,
        threads_per_rank=64,
        launch_params={"cpu_affinity": "depth"},
    )
    j.save()
    return j


def run(
    site_id: int,
    app_id: int,
    project: str,
    queue: str,
    experiment_tag: str,
    num_nodes: int,
    time_limit_min: int,
    jobs_per_sec: float,
    remote_alias: str,
    result_dir: Path,
    matrix_inputs: List[Path],
    size_tag: str,
) -> None:
    global RESULT_IDX
    tags = {
        "experiment": experiment_tag,
        "job_source": remote_alias,
        "num_nodes": str(num_nodes),
        "size": size_tag,
    }
    tag_qs = [f"{k}:{v}" for k, v in tags.items()]
    sleep_time = 1.0 / jobs_per_sec
    start = datetime.utcnow()
    batch_job = BatchJob.objects.create(
        num_nodes=num_nodes,
        wall_time_min=time_limit_min,
        job_mode="mpi",  # type: ignore
        site_id=site_id,
        project=project,
        queue=queue,
        filter_tags=tags,
    )
    print("Created BatchJob", batch_job)
    while datetime.utcnow() - start < timedelta(minutes=time_limit_min + 1):
        time.sleep(sleep_time)
        active_jobs = Job.objects.filter(app_id=app_id, tags=tag_qs, state=ACTIVE_STATES)
        num_active = active_jobs.count()
        assert num_active is not None

        if num_active < num_nodes:
            workdir = Path(experiment_tag) / (str(num_nodes) + str(size_tag)) / f"{RESULT_IDX:06d}"
            inp_path = random.choice(matrix_inputs)
            job = submit_eig_job(
                app_id=app_id,
                workdir=workdir,
                tags=tags,
                remote_alias=remote_alias,
                matrix_inp_path=inp_path,
                result_dir=result_dir,
            )
            print(f"Submitted job {job.id} to process {inp_path.name}")
            RESULT_IDX += 1

    print("Run Done, waiting for BatchJob to finish")
    batch_job.refresh_from_db()
    i = 0
    while batch_job.state != "finished" and i < 6:
        time.sleep(10)
        batch_job.refresh_from_db()
        i += 1
    print(f"Finished waiting on BatchJob (state {batch_job.state}); Moving on to next run")


def main() -> None:
    with open("eig_scaling.yml") as fp:
        conf = Conf(**yaml.safe_load(fp))

    site = Site.objects.get(id=conf.site_id)
    app = App.objects.get(site_id=site.id, class_path=APP_CLASS)
    assert site.id is not None
    assert app.id is not None

    for run_conf in conf.run_list:
        print("Starting experiment:", run_conf)
        run(
            site_id=site.id,
            app_id=app.id,
            project=conf.project,
            queue=conf.queue,
            experiment_tag=conf.experiment_tag,
            **run_conf.dict(),
        )


if __name__ == "__main__":
    main()



# type: ignore
import os
import random
import time

import click
import globus_sdk
from configobj import ConfigObj
from globus_sdk import RefreshTokenAuthorizer, TransferClient
from globus_sdk.exc import AuthAPIError, NetworkError

CLIENT_ID_OPTNAME = "client_id"
CLIENT_SECRET_OPTNAME = "client_secret"
TEMPLATE_ID_OPTNAME = "template_id"
DEFAULT_TEMPLATE_ID = "95fdeba8-fac2-42bd-a357-e068d82ff78e"
TRANSFER_AT_EXPIRES_OPTNAME = "transfer_access_token_expires"
TRANSFER_RT_OPTNAME = "transfer_refresh_token"
TRANSFER_AT_OPTNAME = "transfer_access_token"
AUTH_RT_OPTNAME = "auth_refresh_token"
AUTH_AT_OPTNAME = "auth_access_token"
AUTH_AT_EXPIRES_OPTNAME = "auth_access_token_expires"
SCOPES = (
    "openid profile email "
    "urn:globus:auth:scope:auth.globus.org:view_identity_set "
    "urn:globus:auth:scope:transfer.api.globus.org:all"
)


class RetryingTransferClient(TransferClient):
    """
    Wrapper around TransferClient that retries safe resources on NetworkErrors
    """

    default_retries = 10

    def __init__(self, tries=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tries = tries or self.default_retries

    def retry(self, f, *args, **kwargs):
        """
        Retries the given function self.tries times on NetworkErros
        """
        backoff = random.random() / 100  # 5ms on average
        for _ in range(self.tries - 1):
            try:
                return f(*args, **kwargs)
            except NetworkError:
                time.sleep(backoff)
                backoff *= 2
        return f(*args, **kwargs)

    # get and put should always be safe to retry
    def get(self, *args, **kwargs):
        return self.retry(super().get, *args, **kwargs)

    def put(self, *args, **kwargs):
        return self.retry(super().put, *args, **kwargs)

    # task submission is safe, as the data contains a unique submission-id
    def submit_transfer(self, *args, **kwargs):
        return self.retry(super().submit_transfer, *args, **kwargs)

    def submit_delete(self, *args, **kwargs):
        return self.retry(super().submit_delete, *args, **kwargs)


def _update_tokens(token_response):
    tokens = token_response.by_resource_server["transfer.api.globus.org"]
    set_transfer_tokens(tokens["access_token"], tokens["refresh_token"], tokens["expires_at_seconds"])


def set_transfer_tokens(access_token, refresh_token, expires_at) -> None:
    write_option(TRANSFER_AT_OPTNAME, access_token)
    write_option(TRANSFER_RT_OPTNAME, refresh_token)
    write_option(TRANSFER_AT_EXPIRES_OPTNAME, expires_at)


def get_config_obj(system=False, file_error=False):
    if system:
        path = "/etc/globus.cfg"
    else:
        path = os.path.expanduser("~/.globus.cfg")

    conf = ConfigObj(path, encoding="utf-8", file_error=file_error)

    # delete any old whomai values in the cli section
    for key in conf.get("cli", {}):
        if "whoami_identity_" in key:
            del conf["cli"][key]
            conf.write()

    return conf


def lookup_option(option, section="cli", environment=None):
    conf = get_config_obj()
    try:
        if environment:
            return conf["environment " + environment][option]
        else:
            return conf[section][option]
    except KeyError:
        return None


def write_option(option, value, section="cli", system=False):
    """
    Write an option to disk -- doesn't handle config reloading
    """
    # deny rwx to Group and World -- don't bother storing the returned old mask
    # value, since we'll never restore it in the CLI anyway
    # do this on every call to ensure that we're always consistent about it
    os.umask(0o077)

    # FIXME: DRY violation with config_commands.helpers
    conf = get_config_obj(system=system)

    # add the section if absent
    if section not in conf:
        conf[section] = {}

    conf[section][option] = value
    conf.write()


def get_transfer_tokens():
    expires = lookup_option(TRANSFER_AT_EXPIRES_OPTNAME)
    if expires is not None:
        expires = int(expires)

    return {
        "refresh_token": lookup_option(TRANSFER_RT_OPTNAME),
        "access_token": lookup_option(TRANSFER_AT_OPTNAME),
        "access_token_expires": expires,
    }


def internal_native_client():
    template_id = lookup_option(TEMPLATE_ID_OPTNAME) or DEFAULT_TEMPLATE_ID
    return globus_sdk.NativeAppAuthClient(template_id)


def internal_auth_client(requires_instance=False, force_new_client=False):
    """
    Looks up the values for this CLI's Instance Client in config

    If none exists and requires_instance is True or force_new_client is True,
    registers a new Instance Client with Globus Auth

    If none exists and requires_instance is false, defaults to a Native Client
    for backwards compatibility

    Returns either a NativeAppAuthClient or a ConfidentialAppAuthClient
    """
    client_id = lookup_option(CLIENT_ID_OPTNAME)
    client_secret = lookup_option(CLIENT_SECRET_OPTNAME)
    template_id = lookup_option(TEMPLATE_ID_OPTNAME) or DEFAULT_TEMPLATE_ID
    template_client = internal_native_client()
    existing = client_id and client_secret

    # if we are forcing a new client, delete any existing client
    if force_new_client and existing:
        existing_client = globus_sdk.ConfidentialAppAuthClient(client_id, client_secret)
        try:
            existing_client.delete(f"/v2/api/clients/{client_id}")

        # if the client secret has been invalidated or the client has
        # already been removed, we continue on
        except globus_sdk.exc.AuthAPIError:
            pass

    # if we require a new client to be made
    if force_new_client or (requires_instance and not existing):
        # register a new instance client with auth
        body = {"client": {"template_id": template_id, "name": "Globus CLI"}}
        res = template_client.post("/v2/api/clients", json_body=body)

        # get values and write to config
        credential_data = res["included"]["client_credential"]
        client_id = credential_data["client"]
        client_secret = credential_data["secret"]
        write_option(CLIENT_ID_OPTNAME, client_id)
        write_option(CLIENT_SECRET_OPTNAME, client_secret)

        return globus_sdk.ConfidentialAppAuthClient(client_id, client_secret, app_name="Globus CLI")

    # if we already have a client, just return it
    elif existing:
        return globus_sdk.ConfidentialAppAuthClient(client_id, client_secret, app_name="Globus CLI")

    # fall-back to a native client to not break old logins
    # TOOD: eventually remove this behavior
    else:
        return template_client


def get_client():
    tokens = get_transfer_tokens()
    authorizer = None

    # if there's a refresh token, use it to build the authorizer
    if tokens["refresh_token"] is not None:
        authorizer = RefreshTokenAuthorizer(
            tokens["refresh_token"],
            internal_auth_client(),
            tokens["access_token"],
            tokens["access_token_expires"],
            on_refresh=_update_tokens,
        )

    return RetryingTransferClient(authorizer=authorizer, app_name="Globus CLI v2.1.0")


def exchange_code_and_store_config(auth_client, auth_code):
    """
    Finishes auth flow after code is gotten from command line or local server.
    Exchanges code for tokens and gets user info from auth.
    Stores tokens and user info in config.
    """
    # do a token exchange with the given code
    tkn = auth_client.oauth2_exchange_code_for_tokens(auth_code)
    tkn = tkn.by_resource_server

    store_queue = []

    def _enqueue(optname, newval, revoke=True):
        store_queue.append((optname, newval, revoke))

    # extract access tokens from final response
    if "transfer.api.globus.org" in tkn:
        _enqueue(TRANSFER_RT_OPTNAME, tkn["transfer.api.globus.org"]["refresh_token"])
        _enqueue(TRANSFER_AT_OPTNAME, tkn["transfer.api.globus.org"]["access_token"])
        _enqueue(
            TRANSFER_AT_EXPIRES_OPTNAME,
            tkn["transfer.api.globus.org"]["expires_at_seconds"],
            revoke=False,
        )
    if "auth.globus.org" in tkn:
        _enqueue(AUTH_RT_OPTNAME, tkn["auth.globus.org"]["refresh_token"])
        _enqueue(AUTH_AT_OPTNAME, tkn["auth.globus.org"]["access_token"])
        _enqueue(
            AUTH_AT_EXPIRES_OPTNAME,
            tkn["auth.globus.org"]["expires_at_seconds"],
            revoke=False,
        )

    # revoke any existing tokens
    for optname in [optname for (optname, _val, revoke) in store_queue if revoke]:
        token = lookup_option(optname)
        if token:
            auth_client.oauth2_revoke_token(token)

    # write new data to config
    for optname, newval, _revoke in store_queue:
        write_option(optname, newval)


def do_link_auth_flow(session_params=None, force_new_client=False):
    """
    Prompts the user with a link to authenticate with globus auth
    and authorize the CLI to act on their behalf.
    """
    session_params = session_params or {}

    # get the ConfidentialApp client object
    auth_client = internal_auth_client(requires_instance=True, force_new_client=force_new_client)

    # start the Confidential App Grant flow
    auth_client.oauth2_start_flow(
        redirect_uri=auth_client.base_url + "v2/web/auth-code",
        refresh_tokens=True,
        requested_scopes=SCOPES,
    )

    # prompt
    additional_params = {"prompt": "login"}
    additional_params.update(session_params)
    linkprompt = "Please authenticate with Globus here"
    click.echo(
        "{0}:\n{1}\n{2}\n{1}\n".format(
            linkprompt,
            "-" * len(linkprompt),
            auth_client.oauth2_get_authorize_url(additional_params=additional_params),
        )
    )

    # come back with auth code
    auth_code = click.prompt("Enter the resulting Authorization Code here").strip()

    # finish auth flow
    exchange_code_and_store_config(auth_client, auth_code)
    return True


def check_logged_in():
    # first, try to get the refresh tokens from config
    # we can skip the access tokens and their expiration times as those are not
    # strictly necessary
    transfer_rt = lookup_option(TRANSFER_RT_OPTNAME)
    auth_rt = lookup_option(AUTH_RT_OPTNAME)

    # if either of the refresh tokens are null return False
    if transfer_rt is None or auth_rt is None:
        return False

    # get or create the instance client
    auth_client = internal_auth_client(requires_instance=True)

    # check that tokens and client are valid
    try:
        for tok in (transfer_rt, auth_rt):
            res = auth_client.oauth2_validate_token(tok)
            if not res["active"]:
                return False

    # if the instance client is invalid, an AuthAPIError will be raised
    # we then force a new client to be created before continuing
    except AuthAPIError:
        return False

    return True



from typing import TYPE_CHECKING, Any, Dict, Iterable, Iterator, List, Optional, TypeVar, Union, overload

from .model import BalsamModel

T = TypeVar("T", bound=BalsamModel)

if TYPE_CHECKING:
    from .manager import Manager

    U = TypeVar("U", bound="Query")  # type: ignore

REPR_OUTPUT_SIZE = 20


class Query(Iterable[T]):
    def __init__(self, manager: "Manager[T]") -> None:
        self._manager: "Manager[T]" = manager
        self._result_cache: Optional[List[T]] = None
        self._filters: Dict[str, Any] = {}
        self._order_field: Optional[str] = None
        self._count: Optional[int] = None
        self._limit: Optional[int] = None
        self._offset: Optional[int] = None
        self._empty: bool = False

    def __repr__(self) -> str:
        data = list(self[: REPR_OUTPUT_SIZE + 1])
        if len(data) > REPR_OUTPUT_SIZE:
            data[-1] = "...(remaining elements truncated)..."  # type: ignore
        return "<%s %r>" % (self.__class__.__name__, data)

    def __len__(self) -> int:
        self._fetch_cache()
        assert isinstance(self._result_cache, list)
        return len(self._result_cache)

    def __bool__(self) -> bool:
        self._fetch_cache()
        return bool(self._result_cache)

    def __setitem__(self, k: int, v: T) -> None:
        if not isinstance(k, int):
            raise TypeError("Item assignment only support for int index")

        if self._result_cache is None:
            self._fetch_cache()

        assert isinstance(self._result_cache, list)
        self._result_cache[k] = v

    @overload
    def __getitem__(self, k: int) -> T:
        ...

    @overload
    def __getitem__(self, k: slice) -> Union[List[T], "Query[T]"]:  # noqa: F811
        ...

    def __getitem__(self, k: Union[int, slice]) -> Union[List[T], T, "Query[T]"]:  # noqa: F811
        """
        Retrieve an item or slice from the set of results.
        """
        if not isinstance(k, (int, slice)):
            raise TypeError("Query indices must be integers or slices, not %s." % type(k).__name__)
        assert (not isinstance(k, slice) and (k >= 0)) or (
            isinstance(k, slice) and (k.start is None or k.start >= 0) and (k.stop is None or k.stop >= 0)
        ), "Negative indexing is not supported."

        if self._result_cache is not None:
            return self._result_cache[k]

        if isinstance(k, slice):
            start: Optional[int]
            stop: Optional[int]
            clone = self._clone()
            if k.start is not None:
                start = int(k.start)
            else:
                start = None
            if k.stop is not None:
                stop = int(k.stop)
            else:
                stop = None
            clone._set_limits(start, stop)
            return list(clone)[start : stop : k.step] if k.step else clone
        else:
            clone = self._clone()
            clone._set_limits(k, k + 1)
            clone._fetch_cache()
            assert clone._result_cache is not None
            return clone._result_cache[0]

    @property
    def _is_sliced(self) -> bool:
        return not (self._limit is None and self._offset is None)

    def _clone(self: "U") -> "U":
        clone = type(self)(manager=self._manager)
        clone._filters = self._filters.copy()
        clone._order_field = self._order_field
        clone._limit = self._limit
        clone._offset = self._offset
        clone._empty = self._empty
        return clone

    def _set_limits(self, start: Optional[int], stop: Optional[int]) -> None:
        if start is None:
            start = 0
        self._offset = start
        if stop is not None:
            self._limit = stop - start

    def __iter__(self) -> Iterator[T]:
        self._fetch_cache()
        assert isinstance(self._result_cache, list)
        return iter(self._result_cache)

    def _fetch_cache(self) -> None:
        if self._result_cache is not None:
            return
        if self._empty:
            self._result_cache = []
            return
        instances, count = self._manager._get_list(
            filters=self._filters,
            ordering=self._order_field,
            limit=self._limit,
            offset=self._offset,
        )
        if count is not None:
            self._count = count
        self._result_cache = instances

    def _filter(self: "U", **kwargs: Any) -> "U":
        # TODO: kwargs should expand to the set of filterable model fields
        if self._is_sliced:
            raise AttributeError("Cannot filter a sliced Query")
        clone = self._clone()

        for key, val in kwargs.items():
            if isinstance(val, dict):
                kwargs[key] = [f"{k}:{v}" for k, v in val.items()]
            if isinstance(val, (list, tuple)) and not val:
                clone._empty = True
        clone._filters.update(kwargs)
        return clone

    def _order_by(self: "U", field: Optional[str]) -> "U":
        # TODO: should validate that only order-able fields are accepted
        if self._is_sliced:
            raise AttributeError("Cannot re-order a sliced Query")
        clone = self._clone()
        clone._order_field = field
        return clone

    # Methods that do not return a Query
    # **********************************
    def _get(self, **kwargs: Any) -> T:
        # TODO: kwargs should expand to the set of filterable model fields
        clone: Query[T] = self._filter(**kwargs)
        clone._fetch_cache()
        assert clone._result_cache is not None
        results: List[T] = list(clone)
        nobj = len(results)
        if nobj == 1:
            return results[0]
        elif nobj == 0:
            raise self._manager._model_class.DoesNotExist(clone._filters)
        else:
            raise self._manager._model_class.MultipleObjectsReturned(nobj)

    def first(self) -> T:
        return self[0]

    def count(self) -> Optional[int]:
        if self._empty:
            return 0
        if self._count is None:
            _, _count = self._manager._get_list(filters=self._filters, limit=0, offset=0, ordering=None)
            self._count = _count
        return self._count

    def _update(self, **kwargs: Any) -> List[T]:
        # TODO: kwargs should expand to a set of allowed update_fields
        if self._empty:
            return []
        return self._manager._do_bulk_update_query(patch=kwargs, filters=self._filters)

    def delete(self) -> None:
        if self._empty:
            return
        return self._manager._do_bulk_delete(filters=self._filters)



import logging

from fastapi import FastAPI, HTTPException, Request, WebSocket, status
from fastapi.responses import JSONResponse
from sqlalchemy.orm.exc import NoResultFound

from balsam.server import settings
from balsam.server.utils import TimingMiddleware, setup_logging

from .auth import build_auth_router, user_from_token
from .pubsub import pubsub
from .routers import apps, batch_jobs, events, jobs, sessions, sites, transfers

logger = logging.getLogger("balsam.server.main")

app = FastAPI(
    title="Balsam API",
    version="0.1.0",
)

setup_logging(settings.log_dir, settings.log_level)


@app.exception_handler(NoResultFound)
async def no_result_handler(request: Request, exc: NoResultFound) -> JSONResponse:
    return JSONResponse(
        status_code=status.HTTP_404_NOT_FOUND,
        content={"error": "Not found"},
    )


app.include_router(
    build_auth_router(),
    prefix="/auth",
    tags=["auth"],
    dependencies=[],
)

app.include_router(
    sites.router,
    prefix="/sites",
    tags=["sites"],
    dependencies=[],
    responses={404: {"description": "Not found"}},
)

app.include_router(
    apps.router,
    prefix="/apps",
    tags=["apps"],
    dependencies=[],
    responses={404: {"description": "Not found"}},
)

app.include_router(
    jobs.router,
    prefix="/jobs",
    tags=["jobs"],
    dependencies=[],
    responses={404: {"description": "Not found"}},
)

app.include_router(
    events.router,
    prefix="/events",
    tags=["events"],
    dependencies=[],
    responses={404: {"description": "Not found"}},
)

app.include_router(
    batch_jobs.router,
    prefix="/batch-jobs",
    tags=["batch-jobs"],
    dependencies=[],
    responses={404: {"description": "Not found"}},
)

app.include_router(
    transfers.router,
    prefix="/transfers",
    tags=["transfers"],
    dependencies=[],
    responses={404: {"description": "Not found"}},
)

app.include_router(
    sessions.router,
    prefix="/sessions",
    tags=["sessions"],
    dependencies=[],
    responses={404: {"description": "Not found"}},
)


@app.websocket("/subscribe-user")
async def subscribe_user(websocket: WebSocket) -> None:
    """
    Subscribe to a stream of all Job events for the authenticated user.
    """
    # Accept and receive token
    await websocket.accept()
    token = await websocket.receive_text()
    try:
        user = user_from_token(token)
    except HTTPException:
        await websocket.close(code=status.WS_1008_POLICY_VIOLATION)
        return

    p = await pubsub.subscribe(user.id)
    while True:
        msg = await p.get_message(timeout=1.0)
        if msg and msg["type"] == "message":
            await websocket.send_bytes(msg["data"])


app.add_middleware(TimingMiddleware, router=app.router)
logger.info("Loaded balsam.server.main")
logger.info(settings.serialize_without_secrets())



from fastapi import status
from fastapi.encoders import jsonable_encoder


class BalsamTestClient:
    def __init__(self, client):
        self._client = client

    @property
    def headers(self):
        return self._client.headers

    def check_stat(self, expect_code, response):
        if expect_code is None:
            return
        if isinstance(expect_code, (list, tuple)):
            assert response.status_code in expect_code, response.text
        else:
            assert response.status_code == expect_code, response.text

    def get(self, url, check=status.HTTP_200_OK, **kwargs):
        """GET kwargs become URL query parameters (e.g. /?site=3)"""
        response = self._client.get(url, params=kwargs)
        self.check_stat(check, response)
        return response.json()

    def post_form(self, url, check=status.HTTP_201_CREATED, **kwargs):
        response = self._client.post(url, data=kwargs)
        self.check_stat(check, response)
        return response.json()

    def post(self, url, check=status.HTTP_201_CREATED, **kwargs):
        response = self._client.post(url, json=jsonable_encoder(kwargs))
        self.check_stat(check, response)
        return response.json()

    def bulk_post(self, url, list_data, check=status.HTTP_201_CREATED):
        response = self._client.post(url, json=jsonable_encoder(list_data))
        self.check_stat(check, response)
        return response.json()

    def put(self, url, check=status.HTTP_200_OK, **kwargs):
        response = self._client.put(url, json=jsonable_encoder(kwargs))
        self.check_stat(check, response)
        return response.json()

    def bulk_put(self, url, payload, check=status.HTTP_200_OK, **kwargs):
        response = self._client.put(url, json=payload, params=kwargs)
        self.check_stat(check, response)
        return response.json()

    def patch(self, url, check=status.HTTP_200_OK, **kwargs):
        response = self._client.patch(url, json=jsonable_encoder(kwargs))
        self.check_stat(check, response)
        return response.json()

    def bulk_patch(self, url, list_data, check=status.HTTP_200_OK):
        response = self._client.patch(url, json=jsonable_encoder(list_data))
        self.check_stat(check, response)
        return response.json()

    def delete(self, url, check=status.HTTP_204_NO_CONTENT):
        response = self._client.delete(url)
        self.check_stat(check, response)
        return response.json()

    def bulk_delete(self, url, check=status.HTTP_204_NO_CONTENT, **kwargs):
        response = self._client.delete(url, params=kwargs)
        self.check_stat(check, response)
        return response.json()


def create_site(
    client,
    name="baz",
    path="/foo",
    transfer_locations={},
    check=status.HTTP_201_CREATED,
):
    return client.post(
        "/sites/",
        name=name,
        path=path,
        transfer_locations=transfer_locations,
        check=check,
    )


def create_app(
    client,
    site_id,
    name="SayHello",
    check=status.HTTP_201_CREATED,
    parameters=None,
    transfers=None,
):
    if parameters is None:
        parameters = {
            "name": {"required": True},
            "N": {"required": False, "default": 1},
        }
    if transfers is None:
        transfers = {
            "hello-input": {
                "required": False,
                "direction": "in",
                "local_path": "hello.yml",
                "description": "Input file for SayHello",
            }
        }
    return client.post(
        "/apps/",
        site_id=site_id,
        name=name,
        serialized_class="txt",
        source_code="txt",
        parameters=parameters,
        transfers=transfers,
        check=check,
    )



import os
from pathlib import Path

LOG_DIR = Path(os.environ.setdefault("BALSAM_LOG_DIR", os.getcwd()))
SERVER_PORT = os.environ.get("SERVER_PORT", "8000")

# Workers
workers = 1
timeout = 60
worker_class = "uvicorn.workers.UvicornWorker"
bind = f"0.0.0.0:{SERVER_PORT}"

# Logging
loglevel = "info"
accesslog = (LOG_DIR / "gunicorn.access").as_posix()
errorlog = (LOG_DIR / "gunicorn.error").as_posix()
access_log_format = repr(
    {"remote": "%(h)s", "date": "%(t)s", "request": "%(r)s", "status": "%(s)s", "response_sec": "%(L)s"}
)
logger_class = "balsam.server.gunicorn_logger.RotatingGunicornLogger"
capture_output = True

# Server Mechanics
proc_name = "balsam-server"
pidfile = "gunicorn.pid"
